"""
  Unit tests for domains.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from __future__ import absolute_import
import numpy as np
# Local imports
from . import domains
from ..utils.general_utils import map_to_bounds
from ..utils.base_test_class import BaseTestClass, execute_tests

class DomainBaseTestCase(object):
  """ Unit tests for Base class. """
  # pylint: disable=no-member

  def setUp(self):
    """ Set up. """
    self.domain_obj = None
    self.points = None
    self.non_points = None
    self._child_set_up()

  def _child_set_up(self):
    """ Child set up. """
    raise NotImplementedError('Implement in a child class.')

  def test_instantiation(self):
    """ Tests instantiation of an object. """
    self.report('Testing instantiation of %s class.'%(type(self.domain_obj)))
    self.report('Created object: %s'%(self.domain_obj), 'test_result')

  def test_membership(self):
    """ Testing membership. """
    self.report('Testing membership of %s class.'%(type(self.domain_obj)))
    points_are_in = [self.domain_obj.is_a_member(pt) for pt in self.points]
    non_points_are_not_in = [not self.domain_obj.is_a_member(pt) for pt in
                             self.non_points]
    assert all(points_are_in)
    assert all(non_points_are_not_in)


class EuclideanDomainTestCase(DomainBaseTestCase, BaseTestClass):
  """ Test class for Euclidean Objects. """

  def _child_set_up(self):
    """ Child set up. """
    self.domain_obj = domains.EuclideanDomain([[0, 2.3], [3.4, 8.9], [0.12, 1.0]])
    self.points = [map_to_bounds(np.random.random((self.domain_obj.dim,)),
                                 self.domain_obj.bounds)
                   for _ in range(5)]
    self.non_points = [map_to_bounds(np.random.random((self.domain_obj.dim,)),
                                     np.array([[3.5, 9.8], [-1.0, 1.1], [2.3, 4.5]]))
                       for _ in range(5)]


class IntegralDomainTestCase(DomainBaseTestCase, BaseTestClass):
  """ Test class for IntegralDomain Objects. """

  def _child_set_up(self):
    """ Child set up. """
    self.domain_obj = domains.IntegralDomain([[0, 10], [-10, 100], [45, 78.4]])
    self.points = [[9, 56, 78], [5, 0, 68], [0, -1, 70]]
    self.non_points = [[11, 0, 67], [5.6, 11, 58], [4, 3.0, 70], [9, 56, 67, 9]]


class DiscreteDomainTestCase(DomainBaseTestCase, BaseTestClass):
  """ Discrete Domain. """

  def _child_set_up(self):
    """ Child set up. """
    self.domain_obj = domains.DiscreteDomain(['abc', 5, 6.5, int, 'k'])
    self.points = ['k', type(4), 6.5]
    self.non_points = ['ab', 75.8, 'qwerty', None]


class DiscreteNumericDomainTestCase(DomainBaseTestCase, BaseTestClass):
  """ Discrete Numeric Domain. """

  def _child_set_up(self):
    """ Child set up. """
    self.domain_obj = domains.DiscreteNumericDomain([4, 5, 207.2, -2.3])
    self.points = [207.2, 5, -2.3]
    self.non_points = [5.4, -1.1, 'kky', None]

  def test_non_numeric_discrete_domain(self):
    """ Constructor. """
    self.report('Testing if exception is raised for non numeric elements in a ' +
                'discrete domain.')
    exception_raised = False
    try:
      domains.DiscreteNumericDomain(['abc', 5, 6.5, int, 'k'])
    except ValueError:
      exception_raised = True
    assert exception_raised


class ProdDiscreteDomainTestCase(DomainBaseTestCase, BaseTestClass):
  """ ProdDiscreteDomain Domain. """

  def _child_set_up(self):
    """ Child set up. """
    self.domain_obj = domains.ProdDiscreteDomain([['abc', 5, 6.5], [None, float]])
    self.points = [('abc', float), [6.5, None], [5, None]]
    self.non_points = [['abc', float, float], [5, 7], 'qwerty', [99], 6]


class ProdDiscreteNumericDomain(DomainBaseTestCase, BaseTestClass):
  """ ProdDiscreteDomain Domain. """

  def _child_set_up(self):
    """ Child set up. """
    self.domain_obj = domains.ProdDiscreteNumericDomain(
                        [[4.3, 2.1, 9.8, 10],
                         [11.2, -23.1, 19.8],
                         [1123, 213, 1980, 1.1]])
    self.points = [[2.1, -23.1, 1980], (10, 11.2, 1.1), [9.8, 19.8, 1123]]
    self.non_points = [[2.1 -13.1, 1980], ('kky', 11.2, 1.1), [9.8, 19.8, 1123, 21]]

  def test_non_numeric_prod_discrete_domain(self):
    """ Constructor. """
    self.report('Testing if exception is raised non numeric product discrete domain.')
    exception_raised = False
    try:
      domains.ProdDiscreteNumericDomain(
        [[2.1, 9.8, 10],
         [11.2, -23.1, 19.8],
         [1123, 213, 'squirrel', 1.1]])
    except ValueError:
      exception_raised = True
    assert exception_raised


class CartesianProductDomainTestCase(DomainBaseTestCase, BaseTestClass):
  """ Unit tests for CartesianProductDomain. """

  def _child_set_up(self):
    """ Child set up. """
    list_of_domains = [
      domains.EuclideanDomain([[0, 2.3], [3.4, 8.9], [0.12, 1.0]]),
      domains.IntegralDomain([[0, 10], [-10, 100], [45, 78.4]]),
      domains.ProdDiscreteDomain([['abc', 5, 6.5], [None, float]]),
      domains.DiscreteDomain(['abc', 5, 6.5, int, 'k']),
      ]
    self.domain_obj = domains.CartesianProductDomain(list_of_domains)
    self.points = [
      [[2.2, 3.6, 0.99], [5, 0, 68], ('abc', float), int],
      [[1.2, 6.4, 0.2], [0, -1, 70], [6.5, None], 'k'],
      [[0.2, 7.9, 0.5], [0, -1, 69], ['abc', None], 5],
      ]
    self.non_points = [
      [[2.2, 3.6, 0.99], [5, 0, 68], ('abc', float), float],
      [[1.2, 6.4, 2.2], [0, -1, 70], [6.5, None], 'k'],
      [[0.2, 7.9, 0.5], [0, -1, 69], ['abc', None, float], 5],
      [[3.2, 7.9, 0.5], [-20, -1, 69], ['abc', None, float], 5],
      [[2.2, 3.6, 0.99], [5, 0, 68], ('abc', float), int, (3.2, 8.5)],
      ]


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for cartesian product GPs.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-member

from argparse import Namespace
import numpy as np
import os
from time import time
# Local
from ..exd.cp_domain_utils import load_cp_domain_from_config_file, \
                                get_processed_func_from_raw_func_for_cp_domain, \
                                get_processed_func_from_raw_func_via_config, \
                                load_config_file, sample_from_config_space
from . import cartesian_product_gp as cpgp
from ..test_data.park1_3.park1_3 import park1_3
from ..test_data.park1_3.park1_3_mf import park1_3_mf
from ..test_data.park2_4.park2_4 import park2_4
from ..test_data.park2_4.park2_4_mf import park2_4_mf
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.general_utils import map_to_bounds, get_idxs_from_list_of_lists


def get_test_dataset(domain_file_name, func, X_train, X_test, kernel=None):
  """ An internal function to organise what is being returned. """
  domain, orderings = load_cp_domain_from_config_file(domain_file_name)
  func = get_processed_func_from_raw_func_for_cp_domain(func, domain, \
           orderings.index_ordering, orderings.dim_ordering)
  Y_train = [func(x) for x in X_train]
  Y_test = [func(x) for x in X_test]
  return Namespace(domain_file_name=domain_file_name, X_train=X_train, Y_train=Y_train,
                   X_test=X_test, Y_test=Y_test, kernel=kernel)


def gen_cpgp_test_data():
  """ Generates CPGP test data. """
  file_dir = os.path.dirname(os.path.realpath(__file__))
  test_data_dir = os.path.dirname(file_dir)
  # pylint: disable=too-many-statements
  # pylint: disable=too-many-locals
  # TODO: test against a naive GP in each of these cases
  # Now define the datasets --------------------------------------------------------------
  ret = []
  n_train = 200
  n_test = 300
  # Dataset 1
  domain_file_name = test_data_dir + '/' + 'test_data/example_configs/eg01.json'
  func = lambda x: x[0][0]**2 + 2*x[1][0] * x[0][0] + 3*x[1][0] + 2.1
  bounds = np.array([[-5, 10], [0, 15]])
  X_train = map_to_bounds(np.random.random((n_train, 2)), bounds)
  X_test = map_to_bounds(np.random.random((n_train, 2)), bounds)
  X_train = [[x] for x in X_train]
  X_test = [[x] for x in X_test]
  ret.append(get_test_dataset(domain_file_name, func, X_train, X_test))
  # Dataset 2
  domain_file_name = test_data_dir + '/' + 'test_data/park2_4/config.json'
  func = park2_4
  bounds = np.array([[0, 1], [0, 1], [103, 194], [10, 11]])
  X_tr_num = map_to_bounds(np.random.random((n_train, len(bounds))), bounds)
  X_te_num = map_to_bounds(np.random.random((n_test, len(bounds))), bounds)
  X_tr_disc = np.random.choice(['rabbit', 'dog', 'hamster', 'gerbil', 'ferret'], n_train)
  X_te_disc = np.random.choice(['rabbit', 'dog', 'hamster', 'gerbil', 'ferret'], n_test)
  X_train = [[[x[0], x[1], x[3]], [int(x[2])], [xd]]
             for (x, xd) in zip(X_tr_num, X_tr_disc)]
  X_test = [[[x[0], x[1], x[3]], [int(x[2])], [xd]]
             for (x, xd) in zip(X_te_num, X_te_disc)]
  ret.append(get_test_dataset(domain_file_name, func, X_train, X_test))
  # Dataset 3
  domain_file_name = test_data_dir + '/' + 'test_data/park1_3/config.json'
  func = park1_3
  bounds = np.array([[10, 16], [0, 1], [0, 1]])
  x2_elems = [4, 10, 23, 45, 78, 87.1, 91.8, 99, 75.7, 28.1, 3.141593]
  modify_func = lambda x1, x2: [[x1[0]],
                  [round(x1[1] * 20)/20, round(x1[2] * 20)/20, x2]]
  X_train_1 = map_to_bounds(np.random.random((n_train, len(bounds))), bounds)
  X_train_2 = np.random.choice(x2_elems, n_train)
  X_train = [modify_func(x1, x2) for (x1, x2) in zip(X_train_1, X_train_2)]
  X_test_1 = map_to_bounds(np.random.random((n_test, len(bounds))), bounds)
  X_test_2 = np.random.choice(x2_elems, n_test)
  X_test = [modify_func(x1, x2) for (x1, x2) in zip(X_test_1, X_test_2)]
  ret.append(get_test_dataset(domain_file_name, func, X_train, X_test))
  # Return
  return ret


def gen_cpmfgp_test_data_from_config_file(config_file_name, raw_func,
  num_tr_data, num_te_data):
  """ Generates datasets for CP Multi-fidelity GP fitting. """
  # Generate data
  def _generate_data(_proc_func, _config, _num_data):
    """ Generates data. """
    ZX_proc = sample_from_config_space(_config, _num_data)
    YY_proc = [_proc_func(z, x) for (z, x) in ZX_proc]
    ZZ_proc = get_idxs_from_list_of_lists(ZX_proc, 0)
    XX_proc = get_idxs_from_list_of_lists(ZX_proc, 1)
    return ZZ_proc, XX_proc, YY_proc, ZX_proc
  # Get dataset for testing
  def _get_dataset_for_testing(_proc_func, _config, _num_tr_data, _num_te_data):
    """ Get dataset for testing. """
    ZZ_train, XX_train, YY_train, ZX_train = _generate_data(_proc_func, _config,
                                                            _num_tr_data)
    ZZ_test, XX_test, YY_test, ZX_test = _generate_data(_proc_func, _config, _num_te_data)
    return Namespace(config_file_name=config_file_name, config=config, raw_func=raw_func,
      ZZ_train=ZZ_train, XX_train=XX_train, YY_train=YY_train, ZX_train=ZX_train,
      ZZ_test=ZZ_test, XX_test=XX_test, YY_test=YY_test, ZX_test=ZX_test)
  # Generate the data and return
  config = load_config_file(config_file_name)
  proc_func = get_processed_func_from_raw_func_via_config(raw_func, config)
  return _get_dataset_for_testing(proc_func, config, num_tr_data, num_te_data)


def gen_cpmfgp_test_data(num_tr_data, num_te_data):
  """ Generates data on all functions. """
  file_dir = os.path.dirname(os.path.realpath(__file__))
  test_data_dir = os.path.dirname(file_dir)
  test_problems = [
    (test_data_dir + '/test_data/park1_3/config_mf.json', park1_3_mf),
    (test_data_dir + '/test_data/park2_4/config_mf.json', park2_4_mf),
    ]
  ret = [gen_cpmfgp_test_data_from_config_file(cfn, rf, num_tr_data, num_te_data)
         for cfn, rf in test_problems]
  return ret


def fit_cpgp_with_dataset(dataset, FitterClass):
  """ Fits and returns a CPGP. """
  gp_fitter = FitterClass(dataset.X_train, dataset.Y_train, dataset.domain_file_name)
  ret_gp = gp_fitter.fit_gp()
  return ret_gp[1], gp_fitter.domain


def fit_cpmfgp_with_dataset(dataset, *args, **kwargs):
  """ Fits and returns a CPMFGP. """
  config = dataset.config
  mfgp_fitter = cpgp.CPMFGPFitter(dataset.ZZ_train, dataset.XX_train, dataset.YY_train,
    fidel_space=config.fidel_space, domain=config.domain,
    fidel_space_kernel_ordering=config.fidel_space_orderings.kernel_ordering,
    domain_kernel_ordering=config.domain_orderings.kernel_ordering,
    *args, **kwargs)
  ret = mfgp_fitter.fit_gp()
  return ret[1]


class CPGPTestCaseDefinitions(object):
  """ Defing unit tests for CartesianProduct GPs. """

  @classmethod
  def _compute_rmse(cls, Y_test, preds):
    """ Computes root mean square error. """
    return np.linalg.norm(Y_test - preds)/np.sqrt(len(Y_test))

  def _test_prediction(self, Y_test_pred, Y_test, Y_train):
    """ Tests prediction on the training set. """
    Y_test_pred = np.array(Y_test_pred)
    Y_test = np.array(Y_test)
    Y_train = np.array(Y_train)
    mean_train_err = self._compute_rmse(Y_test, Y_train.mean())
    mean_test_err = self._compute_rmse(Y_test, Y_test.mean())
    mean_gp_pred_error = self._compute_rmse(Y_test, Y_test_pred)
    test_succ = mean_gp_pred_error < mean_train_err
    test_str = 'gp_err=%0.4f, train_mean_err=%0.4f, test_mean_err=%0.4f, succ=%d'%(
      mean_gp_pred_error, mean_train_err, mean_test_err, test_succ)
    return test_succ, test_str

  def test_gp_fitting_and_creation(self):
    """ Tests fitting and creation. """
    self.report('Testing fitting Cartesian Product GP. Probabilistic test, might fail.')
    num_successes = 0
    for idx, dataset in enumerate(self.cpgp_datasets):
      fit_start_time = time()
      fitted_gp, cp_domain = fit_cpgp_with_dataset(dataset, cpgp.CPGPFitter)
      fit_end_time = time()
      fit_time = fit_end_time - fit_start_time
      lml = fitted_gp.compute_log_marginal_likelihood()
      gp_preds, _ = fitted_gp.eval(dataset.X_test, uncert_form='none')
      test_succ, test_str = self._test_prediction(gp_preds, dataset.Y_test,
                                                  dataset.Y_train)
      self.report('[%d/%d] Fitted GP: lml=%0.4f, GP=%s, time_taken=%0.4fs'%(
        idx+1, len(self.cpgp_datasets), lml, fitted_gp, fit_time), 'test_result')
      self.report('   Test result: %s.'%(test_str), 'test_result')
      self.report('   Domain=%s.'%(cp_domain), 'test_result')
      num_successes += test_succ
    success_frac = num_successes/float(self.num_datasets)
    self.report('Num successes = %d/%d (%0.4f).\n'%(num_successes, self.num_datasets,
                success_frac), 'test_result')
    assert success_frac > 0.5


class CPGPTestCase(CPGPTestCaseDefinitions, BaseTestClass):
  """ Test cases for CPGP. """

  def setUp(self):
    """ Set up. """
    self.cpgp_datasets = gen_cpgp_test_data()
    self.num_datasets = len(self.cpgp_datasets)


class CPMFGPTestCaseDefinitions(CPGPTestCaseDefinitions):
  """ Defing unit tests for Multi-fidelity Cartesian Product GPs. """

  def test_gp_fitting_and_creation(self):
    """ Tests fitting and creation. """
    self.report('Testing fitting Multi-fidelity Cartesian Product GPs. ' +
                'Probabilistic test, might fail.')
    num_successes = 0
    for idx, dataset in enumerate(self.cpmfgp_datasets):
      fit_start_time = time()
      fitted_gp = fit_cpmfgp_with_dataset(dataset)
      fit_end_time = time()
      fit_time = fit_end_time - fit_start_time
      lml = fitted_gp.compute_log_marginal_likelihood()
      gp_preds, _ = fitted_gp.eval_at_fidel(dataset.ZZ_test, dataset.XX_test,
                                            uncert_form='none')
      preds_with_eval_func, _ = fitted_gp.eval(dataset.ZX_test, uncert_form='none')
      diff_eval_err = self._compute_rmse(gp_preds, preds_with_eval_func)
      test_succ, test_str = self._test_prediction(gp_preds, dataset.YY_test,
                                                  dataset.YY_train)
      self.report('[%d/%d] Fitted GP: lml=%0.4f, GP=%s, time taken=%0.4fs.'%(
        idx+1, len(self.cpmfgp_datasets), lml, fitted_gp, fit_time), 'test_result')
      self.report('   Test result: %s, diff_eval_err=%0.4f.'%(test_str, diff_eval_err),
                  'test_result')
      self.report('   fidel_space=%s.'%(dataset.config.fidel_space), 'test_result')
      self.report('   domain=%s.'%(dataset.config.domain), 'test_result')
      num_successes += test_succ
      assert diff_eval_err < 1e-5 * np.linalg.norm(gp_preds)
    success_frac = num_successes/float(self.num_datasets)
    self.report('Num successes = %d/%d (%0.4f).\n'%(num_successes, self.num_datasets,
                success_frac), 'test_result')
    assert success_frac > 0.5


class CPMFGPTestCase(CPMFGPTestCaseDefinitions, BaseTestClass):
  """ Unit tests for Multi-fidelity Cartesian Product GPs. """

  def setUp(self):
    """ Set up. """
    num_tr_data = 300
    num_te_data = 300
    self.cpmfgp_datasets = gen_cpmfgp_test_data(num_tr_data, num_te_data)
    self.num_datasets = len(self.cpmfgp_datasets)


if __name__ == '__main__':
  execute_tests()




"""
  Dragonfly: An open source library for scalable and robust Bayesian optimisation.

  This file contains the main APIs. See apis/opt.py and apis/moo.py for function
  definitions and descriptions of the arguments.

  Rerefernce:
    "Tuning Hyperparameters without Grad Students: Scalable and Robust Bayesian
    optimisation with Dragonfly", Kandasamy K, Vysyaraju K V, Neiswanger W, Paria B,
    Collins C R, Schneider J, Poczos B, Xing E P.

  -- kandasamy@cs.cmu.edu
"""

# Main APIs
from .apis.opt import maximise_function, minimise_function, \
                      maximise_multifidelity_function, minimise_multifidelity_function, \
                      maximize_function, minimize_function, \
                      maximize_multifidelity_function, minimize_multifidelity_function
from .apis.moo import multiobjective_maximise_functions, \
                      multiobjective_minimise_functions, \
                      multiobjective_maximize_functions, \
                      multiobjective_minimize_functions

from .exd.cp_domain_utils import load_config_file, load_config




"""
  Acquisition functions for Bayesian Optimisation.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name

from argparse import Namespace
from copy import copy
import numpy as np
from scipy.stats import norm as normal_distro
# Local
from ..utils.general_utils import solve_lower_triangular
from ..gp.gp_core import get_post_covar_from_raw_covar
from ..exd.domains import EuclideanDomain
from ..exd.exd_utils import maximise_with_method

# TODO: implement using different samples for synchronous methods


# Some utilities we will use for all acquisitions below ---------------------------------
def maximise_acquisition(acq_fn, anc_data, *args, **kwargs):
  """ Maximises the acquisition and returns the highest point. acq_fn is the acquisition
      function to be maximised. anc_data is a namespace which contains ancillary data.
  """
  # pylint: disable=unbalanced-tuple-unpacking
  acq_opt_method = anc_data.acq_opt_method
  if anc_data.domain.get_type() == 'euclidean':
    if acq_opt_method in ['rand']:
      acquisition = acq_fn
    else:
      # these methods cannot handle vectorised functions.
      acquisition = lambda x: acq_fn(x.reshape((1, -1)))
  elif anc_data.domain.get_type() == 'cartesian_product':
    # these methods cannot handle vectorised functions.
    acquisition = lambda x: acq_fn([x])
  _, opt_pt = maximise_with_method(acq_opt_method, acquisition, anc_data.domain,
                                   anc_data.max_evals, *args, **kwargs)
  return opt_pt


def _get_gp_eval_for_parallel_strategy(gp, anc_data, uncert_form='std'):
  """ Returns the evaluation function of the gp depending on the parallel strategy and
      the evaluations in progress.
  """
  # 1. With hallucinations
  def _get_halluc_gp_eval(_gp, _halluc_pts, _uncert_form):
    """ Hallucinated GP eval. """
    return lambda x: _gp.eval_with_hallucinated_observations(x, _halluc_pts,
                                                             uncert_form=_uncert_form)
  # Ordinary eval of gp
  def _get_naive_gp_eval(_gp, _uncert_form):
    """ Naive GP eval. """
    return lambda x: _gp.eval(x, uncert_form=_uncert_form)
  # Check parallelisation strategy and return
  if anc_data.handle_parallel == 'halluc' and \
    len(anc_data.eval_points_in_progress) > 0:
    if anc_data.is_mf:
      return _get_halluc_gp_eval(gp, anc_data.eval_fidel_points_in_progress, uncert_form)
    else:
      return _get_halluc_gp_eval(gp, anc_data.eval_points_in_progress, uncert_form)
  else:
    return _get_naive_gp_eval(gp, uncert_form)


def get_gp_sampler_for_parallel_strategy(gp, anc_data):
  """ Returns a function that can draw samples from the posterior gp depending on the
      parallel strategy and the evaluations in progress.
  """
  # 1. With hallucinations
  def _get_halluc_gp_draw_samples(_gp, _halluc_pts):
    """ Hallucinated sampler. """
    return lambda x: _gp.draw_samples_with_hallucinated_observations(1, x,
                                                                     _halluc_pts).ravel()
  def _get_naive_gp_draw_samples(_gp):
    """ Naive sampler. """
    return lambda x: _gp.draw_samples(1, x).ravel()
  # Check parallelisation strategy and return
  if anc_data.handle_parallel == 'halluc' and \
    len(anc_data.eval_points_in_progress) > 0:
    if anc_data.is_mf:
      return _get_halluc_gp_draw_samples(gp, anc_data.eval_fidel_points_in_progress)
    else:
      return _get_halluc_gp_draw_samples(gp, anc_data.eval_points_in_progress)
  else:
    return _get_naive_gp_draw_samples(gp)


def _get_syn_recommendations_from_asy(asy_acq, num_workers, list_of_gps, anc_datas):
  """ Returns a batch of (synchronous recommendations from an asynchronous acquisition.
  """
  def _get_next_and_append(_list_of_objects):
    """ Internal function to return current gp and list of gps. """
    ret = _list_of_objects.pop(0)
    _list_of_objects = _list_of_objects + [ret]
    return ret, _list_of_objects
  # If list_of_gps is not a list, then make it a list.
  if not hasattr(list_of_gps, '__iter__'):
    list_of_gps = [list_of_gps] * num_workers
  if not hasattr(anc_datas, '__iter__'):
    anc_datas = [anc_datas] * num_workers
  # Create copies
  list_of_gps = [copy(gp) for gp in list_of_gps]
  anc_datas = [copy(ad) for ad in anc_datas]
  # Get first recommendation
  next_gp, list_of_gps = _get_next_and_append(list_of_gps)
  next_anc_data, anc_datas = _get_next_and_append(anc_datas)
  recommendations = [asy_acq(next_gp, next_anc_data)]
  for _ in range(1, num_workers):
    next_gp, list_of_gps = _get_next_and_append(list_of_gps)
    next_anc_data, anc_datas = _get_next_and_append(anc_datas)
    next_anc_data.eval_points_in_progress = recommendations
    recommendations.append(asy_acq(next_gp, next_anc_data))
  return recommendations


# Thompson sampling ---------------------------------------------------------------
def asy_ts(gp, anc_data):
  """ Returns a recommendation via TS in the asyuential setting. """
  anc_data = copy(anc_data)
  # Always use a random optimiser with a vectorised sampler for TS.
  if anc_data.acq_opt_method != 'rand':
    anc_data.acq_opt_method = 'rand'
    anc_data.max_evals = 4 * anc_data.max_evals
  gp_sample = get_gp_sampler_for_parallel_strategy(gp, anc_data)
  return maximise_acquisition(gp_sample, anc_data, vectorised=True)

def syn_ts(num_workers, list_of_gps, anc_datas):
  """ Returns a batch of recommendations via TS in the synchronous setting. """
  return _get_syn_recommendations_from_asy(asy_ts, num_workers, list_of_gps, anc_datas)


# Add-UCB --------------------------------------------------------------------------
def _get_add_ucb_beta_th(dim, time_step):
  """ Computes the beta t for UCB based methods. """
  return np.sqrt(0.2 * dim * np.log(2 * dim * time_step + 1))

def _add_ucb(gp, add_kernel, mean_funcs, anc_data):
  """ Common functionality for Additive UCB acquisition under various settings.
  """
  # pylint: disable=undefined-loop-variable
  # prelims
  kernel_list = add_kernel.kernel_list
  groupings = add_kernel.groupings
  total_max_evals = anc_data.max_evals
  kern_scale = add_kernel.hyperparams['scale']
  domain_bounds = anc_data.domain_bounds
  X_train = np.array(gp.X)
  num_groups = len(kernel_list)
  if mean_funcs is None:
    mean_funcs = lambda x: np.array([0] * len(x))
  if not hasattr(mean_funcs, '__iter__'):
    mean_funcs = [mean_funcs] * num_groups
  group_points = []
  num_coordinates = 0
  anc_data.max_evals = total_max_evals//num_groups

  # Now loop through each group
  for group_j, kernel_j, mean_func_j in \
      zip(groupings, kernel_list, mean_funcs):
    # Using a python internal function in a loop is typically a bad idea. But we are
    # using this function only inside this loop, so it should be fine.
    def _add_ucb_acq_j(X_test_j):
      """ Acquisition for the jth group. """
      betath_j = _get_add_ucb_beta_th(len(group_j), anc_data.t)
      X_train_j = X_train[:, group_j]
      K_tetr_j = kern_scale * kernel_j(X_test_j, X_train_j)
      pred_mean_j = K_tetr_j.dot(gp.alpha) + mean_func_j(X_test_j)
      K_tete_j = kern_scale * kernel_j(X_test_j, X_test_j)
      V_j = solve_lower_triangular(gp.L, K_tetr_j.T)
      post_covar_j = K_tete_j - V_j.T.dot(V_j)
      post_covar_j = get_post_covar_from_raw_covar(post_covar_j, gp.noise_var,
                                                   gp.kernel.is_guaranteed_psd())
      post_std_j = np.sqrt(np.diag(post_covar_j))
      return pred_mean_j + betath_j * post_std_j
    # Objatin the jth group
    anc_data_j = copy(anc_data)
    anc_data_j.domain = EuclideanDomain(domain_bounds[group_j])
    point_j = maximise_acquisition(_add_ucb_acq_j, anc_data_j)
    group_points.append(point_j)
    num_coordinates += len(point_j)

  # Now return
  anc_data.max_evals = total_max_evals
  ret = np.zeros((num_coordinates,))
  for point_j, group_j in zip(group_points, groupings):
    ret[group_j] = point_j
  return ret

def asy_add_ucb(gp, anc_data):
  """ Asynchronous Add UCB. """
  return _add_ucb(gp, gp.kernel, None, anc_data)

def syn_add_ucb(num_workers, list_of_gps, anc_datas):
  """ Synchronous Add UCB. """
  # pylint: disable=unused-argument
  return _get_syn_recommendations_from_asy(asy_add_ucb, num_workers, list_of_gps,
                                           anc_datas)

# UCB ------------------------------------------------------------------------------
def _get_gp_ucb_dim(gp):
  """ Returns the dimensionality of the dimension. """
  if hasattr(gp, 'ucb_dim') and gp.ucb_dim is not None:
    return gp.ucb_dim
  elif hasattr(gp.kernel, 'dim'):
    return gp.kernel.dim
  else:
    return 3.0

def _get_ucb_beta_th(dim, time_step):
  """ Computes the beta t for UCB based methods. """
  return np.sqrt(0.5 * dim * np.log(2 * dim * time_step + 1))

def asy_ucb(gp, anc_data):
  """ Returns a recommendation via UCB in the asyuential setting. """
  beta_th = _get_ucb_beta_th(_get_gp_ucb_dim(gp), anc_data.t)
  gp_eval = _get_gp_eval_for_parallel_strategy(gp, anc_data, 'std')
  def _ucb_acq(x):
    """ Computes the GP-UCB acquisition. """
    mu, sigma = gp_eval(x)
    return mu + beta_th * sigma
  return maximise_acquisition(_ucb_acq, anc_data)

def syn_ucb(num_workers, list_of_gps, anc_datas):
  """ Returns a recommendation via Batch UCB in the synchronous setting. """
  return _get_syn_recommendations_from_asy(asy_ucb, num_workers, list_of_gps, anc_datas)

# PI  ----------------------------------------------------------------------------
def asy_pi(gp, anc_data):
  """ Returns a recommendation based on PI (probability of improvement). """
  curr_best = anc_data.curr_max_val
  gp_eval = _get_gp_eval_for_parallel_strategy(gp, anc_data, 'std')
  # PI acquisition with hallucinated observations
  def _pi_acq(x):
    """ Acquisition for GP EI. """
    mu, sigma = gp_eval(x)
    return normal_distro.cdf((mu - curr_best) / sigma)
  return maximise_acquisition(_pi_acq, anc_data)

def syn_pi(num_workers, list_of_gps, anc_datas):
  """ Returns a recommendation via EI in the synchronous setting. """
  return _get_syn_recommendations_from_asy(asy_pi, num_workers, list_of_gps, anc_datas)


# EI  ----------------------------------------------------------------------------
def _expected_improvement_for_norm_diff(norm_diff):
  """ The expected improvement. """
  return norm_diff * normal_distro.cdf(norm_diff) + normal_distro.pdf(norm_diff)

def asy_ei(gp, anc_data):
  """ Returns a recommendation based on GP-EI. """
  curr_best = anc_data.curr_max_val
  gp_eval = _get_gp_eval_for_parallel_strategy(gp, anc_data, 'std')
  # EI acquisition with hallucinated observations
  def _ei_acq(x):
    """ Acquisition for GP EI. """
    mu, sigma = gp_eval(x)
    norm_diff = (mu - curr_best) / sigma
    return sigma * _expected_improvement_for_norm_diff(norm_diff)
  return maximise_acquisition(_ei_acq, anc_data)

def syn_ei(num_workers, list_of_gps, anc_datas):
  """ Returns a recommendation via EI in the synchronous setting. """
  return _get_syn_recommendations_from_asy(asy_ei, num_workers, list_of_gps, anc_datas)


# TTEI ----------------------------------------------------------------------------------
def _ttei(gp_eval, anc_data, ref_point):
  """ Computes the arm that is expected to do best over ref_point. """
  ref_mean, ref_std = gp_eval([ref_point])
  ref_mean = float(ref_mean)
  ref_std = float(ref_std)
  def _tt_ei_acq(x):
    """ Acquisition for TTEI. """
    mu, sigma = gp_eval(x)
    comb_std = np.sqrt(ref_std**2 + sigma**2)
    norm_diff = (mu - ref_mean)/comb_std
    return comb_std * _expected_improvement_for_norm_diff(norm_diff)
  return maximise_acquisition(_tt_ei_acq, anc_data)

def asy_ttei(gp, anc_data):
  """ Top-Two expected improvement. """
  if np.random.random() < 0.5:
    # With probability 1/2, return the EI point
    return asy_ei(gp, anc_data)
  else:
    max_acq_opt_evals = anc_data.max_evals
    anc_data = copy(anc_data)
    anc_data.max_evals = max_acq_opt_evals//2
    ei_argmax = asy_ei(gp, anc_data)
    # Now return the second argmax
    gp_eval = _get_gp_eval_for_parallel_strategy(gp, anc_data, 'std')
    return _ttei(gp_eval, anc_data, ei_argmax)

def syn_ttei(num_workers, list_of_gps, anc_data):
  """ Returns a recommendation via TTEI in the synchronous setting. """
  return _get_syn_recommendations_from_asy(asy_ttei, num_workers, list_of_gps, anc_data)

# Random --------------------------------------------------------------------------------
def asy_rand(_, anc_data):
  """ Returns random values for the acquisition. """
  def _rand_eval(_):
    """ Acquisition for asy_rand. """
    return np.random.random((1,))
  return maximise_acquisition(_rand_eval, anc_data)

def syn_rand(num_workers, list_of_gps, anc_data):
  """ Returns random values for the acquisition. """
  return _get_syn_recommendations_from_asy(asy_rand, num_workers, list_of_gps, anc_data)


# Multi-fidelity Strategies ==============================================================
def _get_fidel_to_opt_gp(mfgp, fidel_to_opt):
  """ Returns a GP for Boca that can be used to optimise the acquisition
      at fidel_to_opt. """
  boca_gp = Namespace()
  boca_gp.eval = lambda x, *args, **kwargs: mfgp.eval_at_fidel([fidel_to_opt] * len(x),
                                                               x, *args, **kwargs)
  boca_gp.eval_with_hallucinated_observations = \
    lambda x, halluc_fidel_pts, *args, **kwargs: mfgp.eval_with_hallucinated_observations(
      mfgp.get_ZX_from_ZZ_XX([fidel_to_opt] * len(x), x), halluc_fidel_pts,
      *args, **kwargs)
  boca_gp.draw_samples = lambda n, x, *args, **kwargs: mfgp.draw_samples(n,
                      mfgp.get_ZX_from_ZZ_XX([fidel_to_opt] * len(x), x), *args, **kwargs)
  boca_gp.draw_samples_with_hallucinated_observations = \
                lambda n, x, halluc_fidel_pts, *args, **kwargs: \
                  mfgp.draw_samples_with_hallucinated_observations(n,
                    mfgp.get_ZX_from_ZZ_XX([fidel_to_opt] * len(x), x), halluc_fidel_pts,
                                           *args, **kwargs)
  boca_gp.kernel = mfgp.get_domain_kernel()
  return boca_gp

def _add_ucb_for_boca(mfgp, fidel_to_opt, mean_funcs, anc_data):
  """ Add UCB for BOCA. """
  # pylint: disable=undefined-loop-variable
  # TODO: this is repeating a lot of code from add_ucb. Fix this!
  # prelims
  domain_kernel_list = mfgp.domain_kernel.kernel_list
  groupings = mfgp.domain_kernel.groupings
  total_max_evals = anc_data.max_evals
  kern_scale = mfgp.kernel.hyperparams['scale']
  domain_bounds = anc_data.domain_bounds
  X_train = np.array(mfgp.XX)
  num_groups = len(domain_kernel_list)
  if mean_funcs is None:
    mean_funcs = lambda x: np.array([0] * len(x))
  if not hasattr(mean_funcs, '__iter__'):
    mean_funcs = [mean_funcs] * num_groups
  group_points = []
  num_coordinates = 0
  anc_data.max_evals = total_max_evals//num_groups
  K_fidel_Z_tr_to_f2o = mfgp.fidel_kernel(mfgp.ZZ, [fidel_to_opt])
  K_fidel_f2o_to_f2o = float(mfgp.fidel_kernel([fidel_to_opt], [fidel_to_opt]))

  # go through each element
  for group_j, kernel_j, mean_func_j in \
    zip(groupings, domain_kernel_list, mean_funcs):
    # Using python internal function here
    def _mf_add_ucb_acq_j(X_test_j):
      """ Acquisition for the j'th group. """
      betath_j = _get_add_ucb_beta_th(len(group_j), anc_data.t)
      X_train_j = X_train[:, group_j]
      K_tetr_domain_j = kernel_j(X_test_j, X_train_j)
      K_tetr_fidel_j = np.repeat(K_fidel_Z_tr_to_f2o.T, len(X_test_j), axis=0)
      K_tetr_j = kern_scale * K_tetr_fidel_j * K_tetr_domain_j
      pred_mean_j = K_tetr_j.dot(mfgp.alpha) + mean_func_j(X_test_j)
      K_tete_j = kern_scale * K_fidel_f2o_to_f2o * kernel_j(X_test_j, X_test_j)
      V_j = solve_lower_triangular(mfgp.L, K_tetr_j.T)
      post_covar_j = K_tete_j - V_j.T.dot(V_j)
      post_covar_j = get_post_covar_from_raw_covar(post_covar_j, mfgp.noise_var,
                                                   mfgp.kernel.is_guaranteed_psd())
      post_std_j = np.sqrt(np.diag(post_covar_j))
      return pred_mean_j + betath_j * post_std_j
    # Objatin the jth group
    anc_data_j = copy(anc_data)
    anc_data_j.domain = EuclideanDomain(domain_bounds[group_j])
    point_j = maximise_acquisition(_mf_add_ucb_acq_j, anc_data_j)

    group_points.append(point_j)
    num_coordinates += len(point_j)

  # return
  anc_data.max_evals = total_max_evals
  ret = np.zeros((num_coordinates,))
  for point_j, group_j in zip(group_points, groupings):
    ret[group_j] = point_j
  return ret

def asy_add_ucb_for_boca(mfgp, fidel_to_opt, anc_data):
  """ Asynchronous Add UCB. """
  return _add_ucb_for_boca(mfgp, fidel_to_opt, None, anc_data)

def syn_add_ucb_for_boca(num_workers, list_of_mfgps, fidel_to_opt, anc_data):
  """ Synchronous Add UCB. """
  # pylint: disable=unused-argument
  raise NotImplementedError('Not Implemented Yet!')

def boca(select_pt_func, mfgp, anc_data, func_caller):
  """ Uses the BOCA strategy to pick the next point and fidelity as described in
      Kandasamy et al. 2017 "Multi-fidelity Bayesian Optimisation with Continuous
      Approximations (https://arxiv.org/pdf/1703.06240.pdf).
      We have some additional heuristics implemented as described in the appendix.
  """
  if anc_data.curr_acq == 'add_ucb':
    next_eval_point = asy_add_ucb_for_boca(mfgp, func_caller.fidel_to_opt, anc_data)
  else:
    fidel_to_opt_gp = _get_fidel_to_opt_gp(mfgp, func_caller.fidel_to_opt)
    next_eval_point = select_pt_func(fidel_to_opt_gp, anc_data)
  candidate_fidels, cost_ratios = func_caller.get_candidate_fidels_and_cost_ratios(
                                    next_eval_point, filter_by_cost=True)
  num_candidates = len(candidate_fidels)
  cost_ratios = np.array(cost_ratios)
  sqrt_cost_ratios = np.sqrt(cost_ratios)
  information_gaps = np.array(func_caller.get_information_gap(candidate_fidels))
  _, cand_fidel_stds = mfgp.eval_at_fidel(candidate_fidels,
                                          [next_eval_point] * num_candidates,
                                          uncert_form='std')
  cand_fidel_stds = cand_fidel_stds / np.sqrt(mfgp.kernel.hyperparams['scale'])
  std_thresholds = anc_data.boca_thresh_coeff * anc_data.y_range * \
                   sqrt_cost_ratios * information_gaps
  high_std_idxs = cand_fidel_stds > std_thresholds
  qualifying_idxs = np.where(high_std_idxs)[0]
  if len(qualifying_idxs) == 0:
    next_eval_fidel = func_caller.fidel_to_opt
  else:
    # If the cost_ratio is larger than a threshold, then just evaluate at fidel to opt.
    qualifying_fidels = [candidate_fidels[idx] for idx in qualifying_idxs]
    # boca 0
    qualifying_sqrt_cost_ratios = sqrt_cost_ratios[qualifying_idxs]
    qualifying_cost_ratios = cost_ratios[qualifying_idxs]
    next_eval_fidel_idx = qualifying_sqrt_cost_ratios.argmin()
    # Select the fidel
    if qualifying_cost_ratios[next_eval_fidel_idx] > \
       anc_data.boca_max_low_fidel_cost_ratio:
      next_eval_fidel = func_caller.fidel_to_opt
    else:
      next_eval_fidel = qualifying_fidels[next_eval_fidel_idx]
  return next_eval_fidel, next_eval_point


# Put all of them into the following namespaces.
syn = Namespace(
  ucb=syn_ucb,
  add_ucb=syn_add_ucb,
  ei=syn_ei,
  pi=syn_pi,
  ttei=syn_ttei,
  ts=syn_ts,
  rand=syn_rand,
  )

asy = Namespace(
  ucb=asy_ucb,
  add_ucb=asy_add_ucb,
  ei=asy_ei,
  pi=asy_pi,
  ttei=asy_ttei,
  ts=asy_ts,
  rand=asy_rand,
  )

seq = Namespace(
  ucb=asy_ucb,
  add_ucb=asy_add_ucb,
  ei=asy_ei,
  pi=asy_pi,
  ttei=asy_ttei,
  ts=asy_ts,
  rand=asy_rand,
  )




"""
  A unit test for function_caller in function_caller and for synthetic
  functions in utils.euclidean_synthetic_functions.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-member

import os
import numpy as np
# Local imports
from . import cp_domain_utils, experiment_caller
from ..utils import euclidean_synthetic_functions as esf
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.ancillary_utils import get_list_of_floats_as_str
# Synthetic functons for CP Domain
from ..test_data.park2_4.park2_4 import park2_4
from ..test_data.park2_4.park2_4_mf import park2_4_mf
from ..test_data.park2_4.park2_4_mf import cost as cost_park2_4
from ..test_data.park1_3.park1_3 import park1_3
from ..test_data.park1_3.park1_3_mf import park1_3_mf
from ..test_data.park1_3.park1_3_mf import cost as cost_park1_3
# from ..test_data.syn_cnn_2.syn_cnn_2 import syn_cnn_2


_TOL = 1e-5


class EuclideanFunctionCallerTestCase(BaseTestClass):
  """ Unit test for synthetic function. """

  def setUp(self):
    """ Prior set up. """
    self.test_function_data = [('hartmann', 3, None, 'no_noise', None),
                               ('hartmann3', None, None, 'no_noise', None),
                               ('hartmann6', None, None, 'no_noise', None),
                               ('branin', None, None, 'no_noise', None),
                               ('borehole', None, None, 'no_noise', None),
                               # With noise
                               ('hartmann6', None, None, 'gauss', 0.1),
                               ('branin', None, None, 'gauss', 0.1),
                               ('borehole', None, None, 'gauss', 10.0),
                               # Multi-fidelity with and without noise
                               ('hartmann3', None, 2, 'no_noise', None),
                               ('branin', None, 3, 'no_noise', None),
                               ('hartmann6', None, 4, 'gauss', 0.1),
                               ('borehole', None, 1, 'gauss', 10.0),
                              ]
    self.num_test_points = 100

  @classmethod
  def _get_test_vals_from_func_caller(cls, func_caller, test_points):
    """ Gets the test points. """
    test_rets = [func_caller.eval_single(elem) for elem in test_points]
    test_vals = [ret[0] for ret in test_rets]
    test_qinfos = [ret[1] for ret in test_rets]
    return test_vals, test_qinfos, test_rets

  @classmethod
  def _get_test_vals_from_mf_func_caller(cls, func_caller, test_points, test_fidels):
    """ Gets the test points. """
    test_rets = [func_caller.eval_at_fidel_single(fidel, point) for \
                 fidel, point in zip(test_fidels, test_points)]
    test_vals = [ret[0] for ret in test_rets]
    test_qinfos = [ret[1] for ret in test_rets]
    return test_vals, test_qinfos, test_rets

  def _test_for_max_val(self, func_caller):
    """ Tests that the optimum value is larger than the test points. """
    test_points = np.random.random((self.num_test_points, func_caller.domain.dim))
    fc_test_vals, _, _ = self._get_test_vals_from_func_caller(func_caller, test_points)
    assert np.all(np.array(fc_test_vals) <= func_caller.maxval)

  @classmethod
  def _test_for_test_err(cls, test_vals_1, test_vals_2, noise_scale):
    """ A single function to do the testing. """
    test_err = np.linalg.norm(np.array(test_vals_1) - np.array(test_vals_2)) / \
               np.sqrt(len(test_vals_1))
    assert test_err <= _TOL + 2 * noise_scale
    return test_err

  def _test_for_func_vals(self, func_caller, unnorm_func_caller):
    """ Tests that the function values correspond to those given. The unit test
        test_all_synthetic_functions calls this for each of the test functions in
        a loop.
    """
    # In this unit test, fc_ denotes for quantities corresponding to the (normalised)
    # func_caller. ufc_ denotes quantities corresponding to the unnorm_func_caller.
    # Test with func_caller and func
    test_points = np.random.random((self.num_test_points, func_caller.domain.dim))
    raw_test_points = func_caller.get_raw_domain_coords(test_points)
    if not func_caller.is_mf():
      raw_test_vals = [func_caller.func(elem) for elem in raw_test_points]
    else:
      test_fidels = np.random.random((self.num_test_points, func_caller.fidel_space.dim))
      raw_test_fidels = func_caller.get_raw_fidel_coords(test_fidels)
      raw_test_vals = [func_caller.func(rfidel, rpoint) for
                       (rfidel, rpoint) in zip(raw_test_fidels, raw_test_points)]
    # First the normalised function caller
    fc_noise_scale = 0.0 if func_caller.noise_scale is None else func_caller.noise_scale
    if not func_caller.is_mf():
      fc_test_vals, _, _ = self._get_test_vals_from_func_caller(func_caller, test_points)
    else:
      fc_test_vals, _, _ = self._get_test_vals_from_mf_func_caller(func_caller,
                             test_points, test_fidels)
    fc_test_err = self._test_for_test_err(raw_test_vals, fc_test_vals, fc_noise_scale)
    # Test with unnormalised function caller
    ufc_noise_scale = 0.0 if func_caller.noise_scale is None else func_caller.noise_scale
    if not func_caller.is_mf():
      ufc_test_vals, _, _ = self._get_test_vals_from_func_caller(unnorm_func_caller,
                                                                 raw_test_points)
    else:
      ufc_test_vals, _, _ = self._get_test_vals_from_mf_func_caller(unnorm_func_caller,
                              raw_test_points, raw_test_fidels)
    ufc_test_err = self._test_for_test_err(raw_test_vals, ufc_test_vals, ufc_noise_scale)
    # Test for normalised and unnormalised callers
    fc_ufc_noise_scale = fc_noise_scale + ufc_noise_scale
    fc_ufc_err = self._test_for_test_err(fc_test_vals, ufc_test_vals, fc_ufc_noise_scale)
    return fc_test_err, ufc_test_err, fc_ufc_err

  def _test_variation_along_fidel_dim(self, func_caller, unnorm_func_caller,
                                      fidel_dim, test_point=None):
    """ Tests and prints variation along the a fidelity dim. """
    _grid_size = 10
    _rem_coords_pre = np.ones((_grid_size, fidel_dim))
    _fidel_dim_coords = np.reshape(np.linspace(0, 1, _grid_size), (_grid_size, 1))
    _rem_coords_post = np.ones((_grid_size, func_caller.fidel_space.dim - fidel_dim - 1))
    fidel_test_grid = np.hstack((_rem_coords_pre, _fidel_dim_coords, _rem_coords_post))
    raw_fidel_test_grid = func_caller.get_raw_fidel_coords(fidel_test_grid)
    test_point = test_point if test_point is not None else \
                 np.random.random((func_caller.domain.dim,))
    raw_test_point = func_caller.get_raw_domain_coords(test_point)
    test_vals_at_grid = [func_caller.eval_at_fidel_single(fidel, test_point)[0] for
                         fidel in fidel_test_grid]
    ufc_test_vals_at_grid = [
      unnorm_func_caller.eval_at_fidel_single(raw_fidel, raw_test_point)[0]
      for raw_fidel in raw_fidel_test_grid]
    test_vals_at_grid_str = get_list_of_floats_as_str(test_vals_at_grid)
    ufc_test_vals_at_grid_str = get_list_of_floats_as_str(ufc_test_vals_at_grid)
    test_point_str = get_list_of_floats_as_str(test_point)
    raw_test_point_str = get_list_of_floats_as_str(raw_test_point)
    self.report('fidel values (normalised func_caller) at x=%s: (fidel_dim=%d) %s'%(
                test_point_str, fidel_dim, test_vals_at_grid_str), 'test_result')
    self.report('fidel values (unnorm_func_caller) at x=%s:  (fidel_dim=%d) %s'%(
                raw_test_point_str, fidel_dim, ufc_test_vals_at_grid_str), 'test_result')

  def test_all_synthetic_functions(self):
    """ Tests all synthetic functions in a loop. """
    for idx, (func_name, domain_dim, fidel_dim, noise_type, noise_scale) in \
      enumerate(self.test_function_data):
      sf_or_mf = 'sf' if fidel_dim is None else 'mf'
      self.report(('Testing %d/%d: %s(%s), domain_dim:%s, fidel_dim:%s, ' +
                   'noise(%s, %s).')%(idx+1, len(self.test_function_data),
                  func_name, sf_or_mf, domain_dim, fidel_dim,
                  noise_type, noise_scale))
      # get the function
      func_caller = esf.get_syn_func_caller(func_name, domain_dim, fidel_dim,
                                            noise_type, noise_scale,
                                            to_normalise_domain=True)
      unnorm_func_caller = esf.get_syn_func_caller(func_name, domain_dim, fidel_dim,
                                                   noise_type, noise_scale,
                                                   to_normalise_domain=False)
      # Test for the maximum values
      if not func_caller.is_noisy():
        self._test_for_max_val(func_caller)
      # Test for the function values and print out result
      fc_err, ufc_err, fc_ufc_err = \
        self._test_for_func_vals(func_caller, unnorm_func_caller)
      self.report(('normalised_err: %0.5f, unnormalised_err: %0.5f., ' +
                   'norm-unnorm_err: %0.5f')%(fc_err, ufc_err, fc_ufc_err), 'test_result')
      # Print variation along a fidelity dimension
      if func_caller.is_mf() and not func_caller.is_noisy():
        self._test_variation_along_fidel_dim(func_caller, unnorm_func_caller, 0)


class CPFunctionCaller(BaseTestClass):
  """ Unit test for Cartesian Product Function Caller. """

  def setUp(self):
    """ set up. """
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.test_function_data = [
      (test_data_pardir + '/test_data/park2_4/config.json', park2_4, 'gauss', 2.1, None),
      (test_data_pardir + '/test_data/park1_3/config.json', park1_3, 'no_noise',
                                                                     None, None),
#       (test_data_pardir + '/test_data/syn_cnn_2/config.json', syn_cnn_2, 'no_noise',
#                                                                       None, None),
      ]
    self.mf_test_function_data = [
      (test_data_pardir + '/test_data/park2_4/config_mf.json', park2_4_mf,
       'gauss', 0.1, cost_park2_4),
      (test_data_pardir + '/test_data/park1_3/config_mf.json', park1_3_mf,
       'gauss', 0.2, cost_park1_3),
      ]
    self.num_test_points = 100

  def test_all_functions(self):
    """ Unit test to test if the function callers are calling correctly. """
    # pylint: disable=too-many-locals
    num_samples = 5
    test_data = self.test_function_data + self.mf_test_function_data
    num_noise_tests = 0.0
    num_noise_successes = 0.0
    for idx, (domain_config_file, raw_func, noise_type, noise_scale,
              raw_fidel_cost_func) in enumerate(test_data):
      config = cp_domain_utils.load_config_file(domain_config_file)
      func_caller = experiment_caller.get_multifunction_caller_from_config(
        raw_func, domain_config_file, raw_fidel_cost_func=raw_fidel_cost_func,
        noise_type=noise_type, noise_scale=noise_scale)
      proc_samples = cp_domain_utils.sample_from_config_space(config, num_samples)
      raw_samples = [cp_domain_utils.get_raw_from_processed_via_config(x, config)
                       for x in proc_samples]
      if hasattr(config, 'fidel_space') and config.fidel_space is not None:
        raw_vals = [raw_func(z, x) for (z, x) in raw_samples]
        fc_vals = [func_caller.eval_at_fidel_single(z, x)[0] for (z, x) in proc_samples]
        # Check if the cost functions are evaluated properly
        raw_fidel_cost_vals = [raw_fidel_cost_func(z) for z, _ in raw_samples]
        fc_fidel_cost_vals = [func_caller.fidel_cost_func(z) for z, _ in proc_samples]
        fidel_cost_err = np.linalg.norm(np.array(raw_fidel_cost_vals) -
                                        np.array(fc_fidel_cost_vals))
        assert fidel_cost_err < 1e-4 * np.linalg.norm(raw_fidel_cost_vals)
        fidel_cost_err_str = ', fidel_cost_err=%0.3f'%(fidel_cost_err)
      else:
        raw_vals = [raw_func(x) for x in raw_samples]
        fc_vals = [func_caller.eval_single(x)[0] for x in proc_samples]
        fidel_cost_err_str = ''
      err = np.linalg.norm(np.array(raw_vals) - np.array(fc_vals))
      noise_scale = 0.0 if noise_type == 'no_noise' else noise_scale
      noise_tolerance = 3 * noise_scale * np.sqrt(num_samples)
      is_succ = err < 1e-4 * np.linalg.norm(np.array(raw_vals)) + noise_tolerance
      if noise_type == 'no_noise':
        assert is_succ
      else:
        num_noise_tests += 1.0
        num_noise_successes += float(is_succ)
      self.report('[%d/%d] Testing functionevals for %s, err=%0.5f, is_succ=%d%s.'%(
        idx + 1, len(test_data), domain_config_file, err, is_succ, fidel_cost_err_str),
        'test_result')
    self.report('num_noise_successes/num_noise_tests = %d/%d.'%(
      num_noise_successes, num_noise_tests), 'test_result')
    assert num_noise_tests == 0 or num_noise_successes/num_noise_tests > 0.75
    self.report('', '')


if __name__ == '__main__':
  execute_tests()




"""
  Harness for Blackbox Experiment Design. Implements a parent class that can be inherited
  by all methods for black-box optimisation.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=super-on-old-class
# pylint: disable=invalid-name
# pylint: disable=redefined-builtin

from __future__ import division
import time
from argparse import Namespace
import numpy as np
# Local imports
from .exd_utils import EVAL_ERROR_CODE
from ..utils.option_handler import get_option_specs
from ..utils.reporters import get_reporter

exd_core_args = [
  get_option_specs('max_num_steps', False, 1e7,
    'If exceeds this many evaluations, stop.'),
  get_option_specs('capital_type', False, 'return_value',
    'Should be one of return_value, cputime, or realtime'),
  get_option_specs('mode', False, 'asy',
    'If \'syn\', uses synchronous parallelisation, else asynchronous.'),
  get_option_specs('build_new_model_every', False, 17,
    'Updates the model via a suitable procedure every this many iterations.'),
  get_option_specs('report_model_on_each_build', False, 0,
    'If True, will report the model each time it is built.'),
  get_option_specs('report_results_every', False, 13,
    'Report results every this many iterations.'),
  # Initialisation
  get_option_specs('init_capital', False, 'default',
    ('The capital to be used for initialisation.')),
  get_option_specs('init_capital_frac', False, None,
    ('The fraction of the total capital to be used for initialisation.')),
  get_option_specs('num_init_evals', False, 20,
    ('The number of evaluations for initialisation. If <0, will use default.')),
  # The amount of effort we will use for initialisation is prioritised by init_capital,
  # init_capital_frac and num_init_evals.
  get_option_specs('prev_evaluations', False, None,
    'Data for any previous evaluations.'),
  get_option_specs('get_initial_qinfos', False, None,
    'A function to obtain initial qinfos.'),
  get_option_specs('init_method', False, 'rand',
    'Method to obtain initial queries. Is used if get_initial_qinfos is None.'),
  ]


mf_exd_args = [
  get_option_specs('fidel_init_method', False, 'rand',
    'Method to obtain initial fidels. Is used if get_initial_qinfos is None.'),
  get_option_specs('init_set_to_fidel_to_opt_with_prob', False, 0.25,
    'Method to obtain initial fidels. Is used if get_initial_qinfos is None.'),
  ]


class ExperimentDesigner(object):
  """ BlackboxExperimenter Class. """
  #pylint: disable=attribute-defined-outside-init
  #pylint: disable=too-many-instance-attributes

  # Methods needed for construction -------------------------------------------------
  def __init__(self, experiment_caller, worker_manager, model=None,
               options=None, reporter=None):
    """ Constructor.
        experiment_caller is an ExperimentCaller instance.
        worker_manager is a WorkerManager instance.
    """
    # Set up attributes.
    self.experiment_caller = experiment_caller
    self.domain = experiment_caller.domain
    self.worker_manager = worker_manager
    self.options = options
    self.reporter = get_reporter(reporter)
    self.model = model
    # Other set up
    self._set_up()

  def _set_up(self):
    """ Some additional set up routines. """
    # Set up some book keeping parameters
    self.available_capital = 0.0
    self.num_completed_evals = 0
    self.step_idx = 0
    self.num_succ_queries = 0
    # Initialise worker manager
    self.worker_manager.set_experiment_designer(self)
    copyable_params_from_worker_manager = ['num_workers']
    for param in copyable_params_from_worker_manager:
      setattr(self, param, getattr(self.worker_manager, param))
    # Other book keeping stuff
    self.last_report_at = 0
    self.last_model_build_at = 0
    self.eval_points_in_progress = []
    self.eval_idxs_in_progress = []
    # Set initial history
    # query infos will maintain a list of namespaces which contain information about
    # the query in send order. Everything else will be saved in receive order.
    self.history = Namespace(query_step_idxs=[],
                             query_points=[],
                             query_send_times=[],
                             query_receive_times=[],
                             query_eval_times=[],
                             query_worker_ids=[],
                             query_qinfos=[],
                             job_idxs_of_workers={k:[] for k in
                                                  self.worker_manager.worker_ids},
                            )
    self.to_copy_from_qinfo_to_history = {
        'step_idx': 'query_step_idxs',
        'point': 'query_points',
        'send_time': 'query_send_times',
        'receive_time': 'query_receive_times',
        'eval_time': 'query_eval_times',
        'worker_id': 'query_worker_ids',
      }
    # Set pre_eval_points and results
    self.prev_eval_points = []
    # Multi-fidelity Set up
    if self.is_an_mf_method() or self.experiment_caller.is_mf():
      self._mf_set_up()
    # Call the child set up.
    self._exd_child_set_up()
    # Post child set up.
    method_prefix = 'asy' if self.is_asynchronous() else 'syn'
    self.full_method_name = method_prefix + '-' + self._get_method_str()
    self.history.full_method_name = self.full_method_name

  def _mf_set_up(self):
    """ Multi-fidelity set up. """
    assert self.experiment_caller.is_mf()
    self.fidel_space = self.experiment_caller.fidel_space
    # Set up history
    self.history.query_fidels = []
    self.history.query_cost_at_fidels = []
    self.to_copy_from_qinfo_to_history['fidel'] = 'query_fidels'
    self.to_copy_from_qinfo_to_history['cost_at_fidel'] = 'query_cost_at_fidels'
    self.eval_fidels_in_progress = []
    # Set up previous evaluations
    self.prev_eval_fidels = []

  def _exd_child_set_up(self):
    """ Set up for a child class of Blackbox Experiment designer. """
    raise NotImplementedError('Implement in a child class of ExperimentDesigner.')

  def _get_method_str(self):
    """ Return a string describing the method. """
    raise NotImplementedError('Implement in a child class of ExperimentDesigner.')

  def _get_problem_str(self):
    """ Return a string describing the method. """
    raise NotImplementedError('Implement in a child class of ExperimentDesigner.')

  def is_asynchronous(self):
    """ Returns true if asynchronous."""
    return self.options.mode.lower().startswith('asy')

  def is_an_mf_method(self):
    """ Returns True if the method is a multi-fidelity method. """
    raise NotImplementedError('Implement in a method implementation.')

  # Book-keeping -------------------------------------------------------------
  def _update_history(self, qinfo):
    """ qinfo is a namespace information about the query. """
    # Update the number of jobs done by each worker regardless
    self.history.job_idxs_of_workers[qinfo.worker_id].append(qinfo.step_idx)
    self.history.query_qinfos.append(qinfo) # Save the qinfo
    # Now store in history
    for qinfo_name, history_name in self.to_copy_from_qinfo_to_history.items():
      attr_list = getattr(self.history, history_name)
      try:
        attr_list.append(getattr(qinfo, qinfo_name))
      except AttributeError:
        attr_list.append('xxx')
    # Do any child update
    self._exd_child_update_history(qinfo)
    # Check for unsuccessful queries
    if qinfo.val != EVAL_ERROR_CODE:
      self.num_succ_queries += 1

  def _exd_child_update_history(self, qinfo):
    """ Any updates to history from the child class. """
    raise NotImplementedError('Implement in a child class of BlackboxExperimenter.')

  def _get_jobs_for_each_worker(self):
    """ Returns the number of jobs for each worker as a list. """
    jobs_each_worker = [len(elem) for elem in
                        list(self.history.job_idxs_of_workers.values())]
    if self.num_workers <= 5:
      jobs_each_worker_str = str(jobs_each_worker)
    else:
      return '[min:%d, max:%d]'%(min(jobs_each_worker), max(jobs_each_worker))
    return  jobs_each_worker_str

  def _get_curr_job_idxs_in_progress(self):
    """ Returns the current job indices in progress. """
    if self.num_workers <= 4:
      return str(self.eval_idxs_in_progress)
    else:
      total_in_progress = len(self.eval_idxs_in_progress)
      min_idx = (-1 if total_in_progress == 0 else min(self.eval_idxs_in_progress))
      max_idx = (-1 if total_in_progress == 0 else max(self.eval_idxs_in_progress))
      dif = -1 if total_in_progress == 0 else max_idx - min_idx
      return '[min:%d, max:%d, dif:%d, tot:%d]'%(min_idx, max_idx, dif, total_in_progress)

  @classmethod
  def _get_multiple_workers_str(cls):
    """ Get string if there are multiple workers. """
    return ''

  def _print_header(self):
    """ Print header. """
    header_str = 'Legend: <iteration_number> (<num_successful_queries>) ' + \
                 'cap=<fraction_of_capital_spent>:: '
    child_header_str = self._get_exd_child_header_str()
    self.reporter.writeln(header_str + child_header_str)

  @classmethod
  def _get_exd_child_header_str(cls):
    """ Get header for child. """
    return ''

  def _report_curr_results(self):
    """ Writes current result to reporter. """
    cap_frac = (np.nan if self.available_capital <= 0 else
                self.get_curr_spent_capital()/self.available_capital)
    report_str = '#%03d (%03d, %0.3f):: '%(self.step_idx, self.num_succ_queries,
                                           cap_frac)
    report_str += self._get_exd_child_report_results_str()
    report_str += self._get_multiple_workers_str()
    self.reporter.writeln(report_str)
    self.last_report_at = self.step_idx

  def _get_exd_child_report_results_str(self):
    """ Returns a string for the specific child method describing the progress. """
    raise NotImplementedError('Implement in a child class of BlackboxExperimenter.')

  # Methods needed for initialisation ----------------------------------------
  def perform_initial_queries(self):
    """ Perform initial queries. """
    # If we already have some pre_eval points then do this.
    # pylint: disable=too-many-branches
    if (hasattr(self.options, 'prev_evaluations') and
        self.options.prev_evaluations is not None):
      self._exd_child_handle_prev_evals()
    else:
      # Set the initial capital
      if self.options.init_capital == 'default':
        self.init_capital = np.clip(5 * self.domain.get_dim(),
                                    max(5.0, 0.025 * self.available_capital),
                                    max(5.0, 0.075 * self.available_capital))
      elif self.options.init_capital is not None:
        self.init_capital = float(self.options.init_capital)
      elif self.options.init_capital_frac is not None:
        self.init_capital = float(self.options.init_capital_frac) * self.available_capital
      else:
        self.init_capital = None
      # Set the function to return the initial qinfos
      if hasattr(self.options, 'get_initial_qinfos') and \
         self.options.get_initial_qinfos is not None:
        get_initial_qinfos = self.options.get_initial_qinfos
      else:
        get_initial_qinfos = self._get_initial_qinfos
      # Send out the initial queries to the worker manager
      if self.init_capital is not None:
        initial_qinfos_w_init_capital = []
        while True:
          if len(initial_qinfos_w_init_capital) == 0:
            if self.options.capital_type == 'return_value':
              num_initial_queries_w_init_capital = int(self.init_capital)
            else:
              num_initial_queries_w_init_capital = int(2 * self.init_capital)
            initial_qinfos_w_init_capital = \
              get_initial_qinfos(num_initial_queries_w_init_capital)
          qinfo = initial_qinfos_w_init_capital.pop(0)
          self.step_idx += 1
          self._wait_for_a_free_worker()
          if self._terminate_initialisation():
            cap_frac = (np.nan if self.available_capital <= 0 else
                        self.get_curr_spent_capital()/self.available_capital)
            self.reporter.writeln('Capital spent on initialisation: %0.4f(%0.4f).'%(
                self.get_curr_spent_capital(), cap_frac))
            break
          self._dispatch_single_experiment_to_worker_manager(qinfo)
      else:
        num_init_evals = int(self.options.num_init_evals)
        if num_init_evals > 0:
          num_init_evals = max(self.num_workers, num_init_evals)
          init_qinfos = get_initial_qinfos(num_init_evals)
          for qinfo in init_qinfos:
            self.step_idx += 1
            self._wait_for_a_free_worker()
            self._dispatch_single_experiment_to_worker_manager(qinfo)

  def _exd_child_handle_prev_evals(self):
    """ Handles pre-evaluations. """
    raise NotImplementedError('Implement in a child class of BlackboxExperimenter.')

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns the initial qinfos. Can be overridden by a child class. """
    # pylint: disable=unused-argument
    # pylint: disable=no-self-use
    return []

  def initialise_capital(self):
    """ Initialises capital. """
    if self.options.capital_type == 'return_value':
      self.spent_capital = 0.0
    elif self.options.capital_type == 'cputime':
      self.init_cpu_time_stamp = time.clock()
    elif self.options.capital_type == 'realtime':
      self.init_real_time_stamp = time.time()

  def get_curr_spent_capital(self):
    """ Computes the current spent time. """
    if self.options.capital_type == 'return_value':
      return self.spent_capital
    elif self.options.capital_type == 'cputime':
      return time.clock() - self.init_cpu_time_stamp
    elif self.options.capital_type == 'realtime':
      return time.time() - self.init_real_time_stamp

  def set_curr_spent_capital(self, spent_capital):
    """ Sets the current spent capital. Useful only in synthetic set ups."""
    if self.options.capital_type == 'return_value':
      self.spent_capital = spent_capital

  def _print_method_description(self):
    """ Prints method description at the beginning of the routine. """
    method_str = self._get_method_str()
    problem_str = self._get_problem_str()
    if self.num_workers > 1:
      parallel_mode = '-asynchronous' if self.is_asynchronous() else '-synchronous'
      method_str += parallel_mode
    self.reporter.writeln('%s with %s using capital %s (%s)'%(problem_str, method_str,
                          str(self.available_capital), self.options.capital_type))

  def run_experiment_initialise(self):
    """ Initialisation before running initialisation. """
    self._print_method_description()
    self.initialise_capital()
    self.perform_initial_queries()
    self._child_run_experiments_initialise()
    self._print_header()

  def _child_run_experiments_initialise(self):
    """ Handles any initialisation before running experiments. """
    raise NotImplementedError('Implement in a child class of BlackboxExperimenter.')

  # Methods needed for querying ----------------------------------------------------
  def _wait_till_free(self, is_free, poll_time):
    """ Waits until is_free returns true. """
    keep_looping = True
    while keep_looping:
      last_receive_time = is_free()
      if last_receive_time is not None:
        # Get the latest set of results and dispatch the next job.
        self.set_curr_spent_capital(last_receive_time)
        latest_results = self.worker_manager.fetch_latest_results()
        for qinfo in latest_results:
          if self.experiment_caller.is_mf() and not hasattr(qinfo, 'cost_at_fidel'):
            qinfo.cost_at_fidel = qinfo.eval_time
          self._update_history(qinfo)
          self._remove_from_in_progress(qinfo)
        self._add_data_to_model(latest_results)
        keep_looping = False
      else:
        time.sleep(poll_time)

  def _wait_for_a_free_worker(self):
    """ Checks if a worker is free and updates with the latest results. """
    self._wait_till_free(self.worker_manager.a_worker_is_free,
                         self.worker_manager.get_poll_time_real())

  def _wait_for_all_free_workers(self):
    """ Checks to see if all workers are free and updates with latest results. """
    self._wait_till_free(self.worker_manager.all_workers_are_free,
                         self.worker_manager.get_poll_time_real())

  def _add_to_in_progress(self, qinfos):
    """ Adds jobs to in progress. """
    for qinfo in qinfos:
      self.eval_idxs_in_progress.append(qinfo.step_idx)
      self.eval_points_in_progress.append(qinfo.point)
      if self.is_an_mf_method():
        self.eval_fidels_in_progress.append(qinfo.fidel)

  def _remove_from_in_progress(self, qinfo):
    """ Removes a job from the in progress status. """
    completed_eval_idx = self.eval_idxs_in_progress.index(qinfo.step_idx)
    self.eval_idxs_in_progress.pop(completed_eval_idx)
    self.eval_points_in_progress.pop(completed_eval_idx)
    if self.is_an_mf_method():
      self.eval_fidels_in_progress.pop(completed_eval_idx)

  def _dispatch_single_experiment_to_worker_manager(self, qinfo):
    """ Dispatches an experiment to the worker manager. """
    # Create a new qinfo namespace and dispatch new job.
    qinfo.send_time = self.get_curr_spent_capital()
    qinfo.step_idx = self.step_idx
    self.worker_manager.dispatch_single_experiment(self.experiment_caller, qinfo)
    self._add_to_in_progress([qinfo])

  def _dispatch_batch_of_experiments_to_worker_manager(self, qinfos):
    """ Dispatches a batch of experiments to the worker manager. """
    for idx, qinfo in enumerate(qinfos):
      qinfo.send_time = self.get_curr_spent_capital()
      qinfo.step_idx = self.step_idx + idx
    self.worker_manager.dispatch_batch_of_experiments(self.func_caller, qinfos)
    self._add_to_in_progress(qinfos)

  # Methods needed for running experiments ---------------------------------------
  def _terminate_now(self):
    """ Returns true if we should terminate now. """
    if self.step_idx >= self.options.max_num_steps:
      self.reporter.writeln('Exceeded %d evaluations. Terminating Now!' %
                            (self.options.max_num_steps))
      return True
    return self.get_curr_spent_capital() >= self.available_capital

  def _terminate_initialisation(self):
    """ Returns true if we should terminate initialization now. """
    return self.get_curr_spent_capital() >= self.init_capital

  def add_capital(self, capital):
    """ Adds capital. """
    self.available_capital += float(capital)

  def _determine_next_query(self):
    """ Determine the next point for evaluation. """
    raise NotImplementedError('Implement in a child class.!')

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determine the next batch of eavluation points. """
    raise NotImplementedError('Implement in a child class.!')

  def _post_process_next_eval_point(self, point):
    """ Post-process the next point for evaluation. By default, returns same point. """
    #pylint: disable=no-self-use
    return point

  def _add_data_to_model(self, qinfos):
    """ Adds data to model. """
    raise NotImplementedError('Implement in a method child class.')

  def _build_new_model(self):
    """ Builds a new model. """
    self.last_model_build_at = self.step_idx
    self._child_build_new_model()
    if self.options.report_model_on_each_build:
      self._report_model()

  def _child_build_new_model(self):
    """ Builds a new model. """
    raise NotImplementedError('Implement in a child class.!')

  def _report_model(self):
    """ Reports the model. Can be overridden by a child class. """
    pass

  def _update_capital(self, qinfos):
    """ Updates the capital according to the cost of the current query. """
    if not hasattr(qinfos, '__iter__'):
      qinfos = [qinfos]
    query_receive_times = []
    for idx in len(qinfos):
      if self.options.capital_type == 'return_value':
        query_receive_times[idx] = qinfos[idx].send_time + qinfos[idx].eval_time
      elif self.options.capital_type == 'cputime':
        query_receive_times[idx] = time.clock() - self.init_cpu_time_stamp
      elif self.options.capital_type == 'realtime':
        query_receive_times[idx] = time.time() - self.init_real_time_stamp
      # Finally add the receive time of the job to qinfo.
      qinfos[idx].receive_time = query_receive_times[idx]
      qinfos[idx].eval_time = qinfos[idx].receive_time - qinfos[idx].send_time
      if qinfos[idx].eval_time < 0:
        raise ValueError(('Something wrong with the timing. send: %0.4f, receive: %0.4f,'
                          + ' eval: %0.4f.') % (qinfos[idx].send_time, \
                                        qinfos[idx].receive_time, qinfos[idx].eval_time))
    # Compute the maximum of all receive times
    max_query_receive_times = max(query_receive_times)
    return max_query_receive_times

  # Methods for running experiments ----------------------------------------------
  def _asynchronous_run_experiment_routine(self):
    """ Optimisation routine for asynchronous part. """
    self._wait_for_a_free_worker()
    qinfo = self._determine_next_query()
    if self.experiment_caller.is_mf() and not hasattr(qinfo, 'fidel'):
      qinfo.fidel = self.experiment_caller.fidel_to_opt
    self._dispatch_single_experiment_to_worker_manager(qinfo)
    self.step_idx += 1

  def _synchronous_run_experiment_routine(self):
    """ Optimisation routine for the synchronous part. """
    self._wait_for_all_free_workers()
    qinfos = self._determine_next_batch_of_queries(self.num_workers)
    self._dispatch_batch_of_experiments_to_worker_manager(qinfos)
    self.step_idx += self.num_workers

  def _run_experiment_wrap_up(self):
    """ Wrap up before concluding running experiments. """
    self.worker_manager.close_all_queries()
    self._wait_for_all_free_workers()
    self._report_curr_results()
    # Store additional data
    self.history.num_jobs_per_worker = np.array(self._get_jobs_for_each_worker())

  def _main_loop_pre(self):
    """ Anything to be done before each iteration of the main loop. Mostly in case
        this is needed by a child class. """
    pass

  def _main_loop_post(self):
    """ Anything to be done after each iteration of the main loop. Mostly in case
        this is needed by a child class. """
    pass

  # run_experiments method --------------
  def run_experiments(self, max_capital):
    """ This is the main loop which executes the experiments in a loop. """
    self.add_capital(max_capital)
    self.run_experiment_initialise()

    # Main loop --------------------
    while not self._terminate_now():
      # Main loop pre
      self._main_loop_pre()
      if self.step_idx - self.last_report_at >= self.options.report_results_every:
        self._report_curr_results()
      # Experimentation step
      if self.is_asynchronous():
        self._asynchronous_run_experiment_routine()
      else:
        self._synchronous_run_experiment_routine()
      # Book keeeping
      if self.step_idx - self.last_model_build_at >= self.options.build_new_model_every:
        self._build_new_model()
      # Main loop post
      self._main_loop_post()

    # Wrap up and return
    self._run_experiment_wrap_up()
    return self._get_final_return_quantities()

  def _get_final_return_quantities(self):
    """ Gets the final quantities to be returned. Can be overriden by a child. """
    return self.history




"""
  Harness for conducting black box optimisation evaluations.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=maybe-no-member

from argparse import Namespace
from datetime import datetime
import os
import numpy as np
# Local imports
from .blackbox_optimiser import OptInitialiser
from ..utils.method_evaluator import BaseMethodEvaluator


class OptMethodEvaluator(BaseMethodEvaluator):
  """ Base class for evaluating methods for optimisation. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, study_name, func_caller, worker_manager, max_capital, methods,
               num_trials, save_dir, evaluation_options, save_file_prefix='',
               method_options=None, reporter=None, **kwargs):
    """ Constructor. Also see BasicExperimenter for more args. """
    # pylint: disable=too-many-arguments
    self.func_caller = func_caller
    save_file_name = self._get_save_file_name(save_dir, study_name, \
                     worker_manager.num_workers, save_file_prefix, \
                     worker_manager.get_time_distro_info(), max_capital)
    super(OptMethodEvaluator, self).__init__(study_name, num_trials,
                                             save_file_name, reporter=reporter, **kwargs)
    self.worker_manager = worker_manager
    self.max_capital = float(max_capital)
    # Methods
    self.methods = methods
    self.num_methods = len(methods)
    self.domain = func_caller.domain
    self.method_options = (method_options if method_options else
                           {key: None for key in methods})
    # Experiment options will have things such as if the evaluations are noisy,
    # the time distributions etc.
    self.evaluation_options = evaluation_options
    self._set_up_saving()

  def _get_save_file_name(self, save_dir, study_name, num_workers, save_file_prefix,
                          time_distro_str, max_capital):
    """ Gets the save file name. """
    save_file_prefix = save_file_prefix if save_file_prefix else study_name
    noisy_str = 'noisy' if self.func_caller.is_noisy() else 'noiseless'
    save_file_name = '%s-%s-M%d-%s-c%d-%s.mat'%(save_file_prefix, noisy_str, num_workers,
      time_distro_str, int(max_capital), datetime.now().strftime('%m%d-%H%M%S'))
    save_file_name = os.path.join(save_dir, save_file_name)
    return save_file_name

  def _set_up_saving(self):
    """ Runs some routines to set up saving. """
    # Store methods and the options in to_be_saved
    self.to_be_saved.max_capital = self.max_capital
    self.to_be_saved.num_workers = self.worker_manager.num_workers
    self.to_be_saved.methods = self.methods
    self.to_be_saved.method_options = self.method_options # Some error created here.
    self.to_be_saved.time_distro_str = self.worker_manager.get_time_distro_info()
    # Data about the problem
    self.to_be_saved.true_maxval = (self.func_caller.maxval
                                    if self.func_caller.maxval is not None else -np.inf)
    self.to_be_saved.true_argmax = (self.func_caller.argmax \
                                  if self.func_caller.argmax is not None else 'not-known')
    self.to_be_saved.domain_type = self.domain.get_type()
    self.to_be_saved.is_noisy = self.func_caller.is_noisy()
    # For the results
    self.data_to_be_saved = ['query_step_idxs',
                             'query_points',
                             'query_vals',
                             'query_true_vals',
                             'query_send_times',
                             'query_receive_times',
                             'query_eval_times',
                             'query_worker_ids',
                             'curr_opt_vals',
                             'curr_opt_points',
                             'curr_true_opt_vals',
                             'curr_true_opt_points',
                             'num_jobs_per_worker']
    self.data_to_be_saved_if_available = ['query_fidels',
                                          'query_cost_at_fidels',
                                          'query_at_fidel_to_opts',
                                          'prev_eval_points',
                                          'prev_eval_vals']
    self.data_not_to_be_mat_saved.extend(['method_options', 'query_points',
                                          'curr_opt_points', 'curr_true_opt_points'])
    self.data_not_to_be_pickled.extend(['method_options'])
    for data_type in self.data_to_be_saved:
      setattr(self.to_be_saved, data_type, self._get_new_empty_results_array())
    for data_type in self.data_to_be_saved_if_available:
      setattr(self.to_be_saved, data_type, self._get_new_empty_results_array())

  def _get_new_empty_results_array(self):
    """ Returns a new empty arrray to be used for saving results. """
#     return np.empty((self.num_methods, 0), dtype=np.object)
    return np.array([[] for _ in range(self.num_methods)], dtype=np.object)

  def _get_new_iter_results_array(self):
    """ Returns an empty array to be used for saving results of current iteration. """
#     return np.empty((self.num_methods, 1), dtype=np.object)
    return np.array([['-'] for _ in range(self.num_methods)], dtype=np.object)

  def _print_method_header(self, full_method_name):
    """ Prints a header for the current method. """
    trial_header = '-- Exp %d/%d on %s:: %s with cap %0.4f. ----------------------'%( \
      self.trial_iter, self.num_trials, self.study_name, full_method_name, \
      self.max_capital)
    self.reporter.writeln(trial_header)

  def get_iteration_header(self):
    """ Header for iteration. """
    noisy_str = ('no-noise' if self.func_caller.noise_type == 'no_noise' else
                 'noisy(%0.2f)'%(self.func_caller.noise_scale))
    maxval_str = ('?' if self.func_caller.maxval is None
                  else '%0.5f'%(self.func_caller.maxval))
    ret = '%s (M=%d), td: %s, max=%s, max-capital %0.2f, %s'%(self.study_name, \
      self.worker_manager.num_workers, self.to_be_saved.time_distro_str, maxval_str, \
      self.max_capital, noisy_str)
    return ret

  def _print_method_result(self, method, comp_opt_val, num_evals):
    """ Prints the result for this method. """
    result_str = 'Method: %s achieved max-val %0.5f in %d evaluations.\n'%(method, \
                  comp_opt_val, num_evals)
    self.reporter.writeln(result_str)

  def _get_prev_eval_qinfos(self):
    """ Gets the initial qinfos for all methods. """
    if self.evaluation_options.prev_eval_points.lower() == 'generate':
      init_pool_qinfos = self._get_initial_pool_qinfos()
    elif self.evaluation_options.prev_eval_points.lower() == 'none':
      return None
    else:
      # Load from the file.
      raise NotImplementedError('Not written reading results from file yet.')
    # Create an initialiser
    initialiser = OptInitialiser(self.func_caller, self.worker_manager)
    initialiser.options.get_initial_qinfos = lambda _: init_pool_qinfos
    initialiser.options.max_num_steps = 0
    _, _, init_hist = initialiser.initialise()
    return init_hist.query_qinfos

  def _get_initial_pool_qinfos(self):
    """ Returns the intial pool to bootstrap methods in one evaluation. """
    raise NotImplementedError('Implement in a child class.')

  def _optimise_with_method_on_func_caller(self, method, func_caller, worker_manager,
                                           max_capital, meth_options, reporter):
    """ Run method on the function caller and return. """
    raise NotImplementedError('Implement in a child class!')

  def run_trial_iteration(self):
    """ Runs each method in self.methods once and stores the results to be saved. """
    curr_iter_results = Namespace()
    for data_type in self.data_to_be_saved:
      setattr(curr_iter_results, data_type, self._get_new_iter_results_array())
    for data_type in self.data_to_be_saved_if_available:
      setattr(curr_iter_results, data_type, self._get_new_iter_results_array())

    # Fetch pre-evaluation points.
    self.worker_manager.reset()
    prev_eval_qinfos = self._get_prev_eval_qinfos()
    if prev_eval_qinfos is not None:
      prev_eval_vals = [qinfo.val for qinfo in prev_eval_qinfos]
      self.reporter.writeln('Using %d pre-eval points with values. eval: %s (%0.4f).'%( \
                            len(prev_eval_qinfos), prev_eval_vals, max(prev_eval_vals)))
    else:
      self.reporter.writeln('Not using any pre-eval points.')


    # Will go through each method in this loop.
    for meth_iter in range(self.num_methods):
      curr_method = self.methods[meth_iter]
      curr_meth_options = self.method_options[curr_method]
      # Set prev_eval points and vals
      if prev_eval_qinfos is not None:
        curr_meth_options.prev_evaluations = Namespace(qinfos=prev_eval_qinfos)
      else:
        curr_meth_options.prev_evaluations = None
      # Reset worker manager
      self.worker_manager.reset()
      self.reporter.writeln( \
        '\nResetting worker manager: worker_manager.experiment_designer:%s'%( \
        str(self.worker_manager.experiment_designer)))

      # Call the method here.
      self._print_method_header(curr_method)
      history = self._optimise_with_method_on_func_caller(curr_method, self.func_caller, \
                  self.worker_manager, self.max_capital, curr_meth_options, self.reporter)

      # Now save results for current method
      for data_type in self.data_to_be_saved:
        data = getattr(history, data_type)
        data_pointer = getattr(curr_iter_results, data_type)
        data_pointer[meth_iter, 0] = data
      for data_type in self.data_to_be_saved_if_available:
        if hasattr(history, data_type):
          data = getattr(history, data_type)
        else:
          data = ['xx'] * len(history.query_points)
        data_pointer = getattr(curr_iter_results, data_type)
        data_pointer[meth_iter, 0] = data
      # Print out results
      comp_opt_val = history.curr_true_opt_vals[-1]
      num_evals = len(history.curr_true_opt_vals)
      self._print_method_result(curr_method, comp_opt_val, num_evals)
      # Save results of current iteration
      self.update_to_be_saved(curr_iter_results)
      self.save_pickle()
      self.save_results()
    # Save here
    self.update_to_be_saved(curr_iter_results)
    self.save_pickle()
    # No need to explicitly save_results() here - it is done by the parent class.

  def update_to_be_saved(self, curr_iter_results):
    """ Updates the results of the data to be saved with curr_iter_results."""
    for data_type in self.data_to_be_saved + self.data_to_be_saved_if_available:
      data = getattr(curr_iter_results, data_type)
      curr_data_to_be_saved = getattr(self.to_be_saved, data_type)
      if curr_data_to_be_saved.shape[1] == self.trial_iter:
        updated_data_to_be_saved = curr_data_to_be_saved
        updated_data_to_be_saved[:, -1] = data.ravel()
      elif curr_data_to_be_saved.shape[1] < self.trial_iter:
        updated_data_to_be_saved = np.concatenate((curr_data_to_be_saved, data), axis=1)
      else:
        raise ValueError('Something wrong with data saving.')
      setattr(self.to_be_saved, data_type, updated_data_to_be_saved)




"""
  Test cases for functions/classes in nn_comparator.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name

import numpy as np
# Local imports
from . import otmann
from .unittest_neural_network import generate_cnn_architectures, \
                                     generate_mlp_architectures
from ..utils.ancillary_utils import get_list_of_floats_as_str
from ..utils.base_test_class import BaseTestClass, execute_tests

_TOL = 1e-5


class TransportNNDistanceComputerTestCase(BaseTestClass):
  """ Contains unit tests for the TransportNNDistanceComputer class. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(TransportNNDistanceComputerTestCase, self).__init__(*args, **kwargs)
    self.non_assignment_penalty = 1
    cnn_layer_labels, label_mismatch_penalty = \
      otmann.get_cnn_layer_label_mismatch_penalties(self.non_assignment_penalty)
    self.tp_comp = otmann.OTMANNDistanceComputer(cnn_layer_labels,
                     label_mismatch_penalty, self.non_assignment_penalty,
                     otmann.CNN_STRUCTURAL_PENALTY_GROUPS,
                     otmann.PATH_LENGTH_TYPES,
                     dflt_mislabel_coeffs=1.0, dflt_struct_coeffs=1.0,
                     dflt_dist_type='lp-emd')
    self.cnns = generate_cnn_architectures()

  def test_cnn_label_mismatch_penalties(self):
    """ Unit test for the label mismatch penalty of a CNN. """
    self.report('Testing generation of label mismatches for a CNN. ')
    cnn_layer_labels, label_mismatch_penalty = \
      otmann.get_cnn_layer_label_mismatch_penalties(self.non_assignment_penalty,
                       max_conv_size=9)
    self.report('cnn_layer_labels: %s'%(str(cnn_layer_labels)), 'test_result')
    self.report('cnn mismatch penalties: \n%s'%(str(np.round(label_mismatch_penalty, 3))),
                                            'test_result')
    assert np.all(label_mismatch_penalty == label_mismatch_penalty.T)
    assert np.all(np.diag(label_mismatch_penalty) == 0)

  def test_mlp_label_mismatch_penalties(self):
    """ Unit test for the label mismatch penalty of an MLP. """
    self.report('Testing generation of label mismatches for a MLP. ')
    mlp_layer_labels, label_mismatch_penalty = \
      otmann.get_mlp_layer_label_mismatch_penalties(self.non_assignment_penalty,
                                                            'reg')
    self.report('mlp_layer_labels: %s'%((str(mlp_layer_labels))), 'test_result')
    self.report('mlp mismatch penalties: \n%s'%(str(np.round(label_mismatch_penalty, 3))),
                'test_result')
    assert np.all(label_mismatch_penalty == label_mismatch_penalty.T)
    assert np.all(np.diag(label_mismatch_penalty) == 0)

  @classmethod
  def _is_cost_matrix_for_same_networks(cls, cost_matrix):
    """ Returns true if it is the cost matrix for the same network. """
    return np.all(np.diag(cost_matrix) == 0) and np.all(cost_matrix == cost_matrix.T)

  @classmethod
  def _has_corresponding_layers(cls, cost_matrix):
    """ Returns true if one network has a corresponding layer in the other and vice
        versa. """
    ret = True
    for row_idx in range(cost_matrix.shape[0]):
      ret = ret and np.any(cost_matrix[row_idx, :] == 0)
    for col_idx in range(cost_matrix.shape[1]):
      ret = ret and np.any(cost_matrix[:, col_idx] == 0)
    return ret

  def test_mislabel_cost_matrix(self):
    """ Tests the mislabel cost matrix for specific pairs of neural networks. """
    self.report('Testing generation of label cost matrices for specific cnns. ')
    num_cnns = len(self.cnns)
    for i in range(num_cnns):
      for j in range(i+1, num_cnns):
        cnn_i = self.cnns[i]
        cnn_j = self.cnns[j]
        mislabel_cost_matrix = self.tp_comp.get_mislabel_cost_matrix(cnn_i, cnn_j)
        assert mislabel_cost_matrix.shape[0] == cnn_i.num_layers
        assert mislabel_cost_matrix.shape[1] == cnn_j.num_layers
        if i == j:
          assert self._is_cost_matrix_for_same_networks(mislabel_cost_matrix)
        if i == 0 and j == 1:
          # These two matrices were designed so that there has to be a zero for each
          # column on each row and vice versa.
          assert self._has_corresponding_layers(mislabel_cost_matrix)
        if (i == 2 and j == 3) or (i == 0 and j == 1) or (i == 1 and j == 6):
          self.report('Mislabel cost matrix for cnn-%d and cnn-%d:\n%s'%(i, j,
                      str(np.round(mislabel_cost_matrix, 3))), 'test_result')

  def test_connectivity_cost_matrix(self):
    """ Tests the connectivity cost matrix for specific pairs of neural networks.  """
    self.report('Testing generation of connectivity cost matrices for specific cnns.')
    num_cnns = len(self.cnns)
    for i in range(num_cnns):
      for j in range(i, num_cnns):
        cnn_i = self.cnns[i]
        cnn_j = self.cnns[j]
        struct_cost_matrix = self.tp_comp.get_struct_cost_matrix(cnn_i, cnn_j)
        assert struct_cost_matrix.shape[0] == cnn_i.num_layers
        assert struct_cost_matrix.shape[1] == cnn_j.num_layers
        if i == j:
          assert self._is_cost_matrix_for_same_networks(struct_cost_matrix)
        if i == 0 and j == 1:
          # These two matrices were designed so that there has to be a zero for each
          # column on each row and vice versa.
          assert self._has_corresponding_layers(struct_cost_matrix)
        if (i == 2 and j == 3) or (i == 0 and j == 1) or (i == 1 and j == 6):
          self.report('Structural cost matrix for cnn-%d and cnn-%d:\n%s'%(i, j,
                      str(np.round(struct_cost_matrix, 3))), 'test_result')

  def test_ot_cost_matrix(self):
    """ Tests the OT cost matrix for specific pairs of neural networks.  """
    self.report('Testing generation of OT cost matrices for specific cnns.')
    nns = self.cnns
    num_nns = len(nns)
    for i in range(num_nns):
      for j in range(i, num_nns):
        nn_i = nns[i]
        nn_j = nns[j]
        mislabel_cost_matrix = self.tp_comp.get_mislabel_cost_matrix(nn_i, nn_j)
        struct_cost_matrix = self.tp_comp.get_struct_cost_matrix(nn_i, nn_j)
        ot_cost_matrix = self.tp_comp.get_ot_cost_matrix(mislabel_cost_matrix,
                           struct_cost_matrix, 1, 0.1, 1, None)
        assert ot_cost_matrix.shape[0] == nn_i.num_layers + 1
        assert ot_cost_matrix.shape[1] == nn_j.num_layers + 1
        if i == j:
          assert self._is_cost_matrix_for_same_networks(ot_cost_matrix)
        if i == 0 and j == 1:
          # These two matrices were designed so that there has to be a zero for each
          # column on each row and vice versa.
          assert self._has_corresponding_layers(ot_cost_matrix)
        if (i == 2 and j == 3) or (i == 0 and j == 1) or (i == 1 and j == 6):
          self.report('OT cost matrix for cnn-%d and cnn-%d:\n%s'%(i, j,
                      str(np.round(ot_cost_matrix, 3))), 'test_result')

  @classmethod
  def _get_dist_type_abbr(cls, dist_type):
    """ Shortens distance type. """
    if dist_type == 'lp-norm-by-max':
      return 'lnbm'
    else:
      return dist_type

  def _test_dist_comp_for_single_conn_coeff(self, nns, dist_types, tp_comp):
    """ Tests distance computation for a single connectivity coefficient. """
    num_nns = len(nns)
    for i in range(num_nns):
      for j in range(i, num_nns):
        nn_i = nns[i]
        nn_j = nns[j]
        dists = {}
        for dt in dist_types:
          dists[dt] = tp_comp(nn_i, nn_j, dist_type=dt)
        res_str = ' '.join(['%s=%s'%(self._get_dist_type_abbr(key),
                                     get_list_of_floats_as_str(val))
                            for key, val in dists.items()])
        self.report('(i,j)=(%d,%d) %s'%(i, j, res_str), 'test_result')

  def _test_dist_comp_for_multiple_coeffs(self, nns, dist_types, mislabel_coeffs,
                                          struct_coeffs, tp_comp):
    """ Tests distance computation for a single connectivity coefficient. """
    num_nns = len(nns)
    for i in range(num_nns):
      for j in range(i, num_nns):
        nn_i = nns[i]
        nn_j = nns[j]
        dists = {}
        for dt in dist_types:
          dists[dt] = tp_comp(nn_i, nn_j, dist_type=dt, mislabel_coeffs=mislabel_coeffs,
                              struct_coeffs=struct_coeffs)
          num_dists = len(dt.split('-')) * len(mislabel_coeffs)
          assert len(dists[dt]) == num_dists
        res_str = ' '.join(['%s=%s'%(self._get_dist_type_abbr(key),
                                     get_list_of_floats_as_str(val))
                            for key, val in dists.items()])
        self.report('(i,j)=(%d,%d) %s'%(i, j, res_str), 'test_result')

  def test_cnn_distance_computation(self):
    """ Tests the computation of the distance for CNNs. """
    struct_coeffs = [0.1, 0.4]
    mislabel_coeffs = [1.0] * len(struct_coeffs)
    dist_types = ['lp-emd', 'lp', 'emd']
    self.report('Testing distance computation for specific cnns with default coeffs.')
    self._test_dist_comp_for_single_conn_coeff(self.cnns, dist_types, self.tp_comp)
    # With multiple conn_coeffs
    self.report('Testing distance computation for specific cnns with conn_coeffs=%s.'%(
                struct_coeffs))
    self._test_dist_comp_for_multiple_coeffs(self.cnns, dist_types, mislabel_coeffs,
                                             struct_coeffs, self.tp_comp)

  def test_mlp_distance_computation(self):
    """ Tests the computation of the distance for CNNs. """
    self.report('Testing distance computation for specific mlps.')
    # Create the transport computation object
    mlp_layer_labels, label_mismatch_penalty = \
      otmann.get_mlp_layer_label_mismatch_penalties(self.non_assignment_penalty,
                                                            'reg')
    mlp_tp_comp = otmann.OTMANNDistanceComputer(mlp_layer_labels,
                     label_mismatch_penalty,
                     self.non_assignment_penalty,
                     otmann.MLP_STRUCTURAL_PENALTY_GROUPS,
                     otmann.PATH_LENGTH_TYPES,
                     dflt_mislabel_coeffs=1.0, dflt_struct_coeffs=1.0)
    # Create the mlp architectures
    mlps = generate_mlp_architectures()
    struct_coeffs = [0.1, 0.2]
    dist_types = ['lp-emd', 'lp', 'emd']
    mislabel_coeffs = [1.0] * len(struct_coeffs)
    self.report('Testing distance computation for specific mlps with default coeffs.')
    self._test_dist_comp_for_single_conn_coeff(mlps, dist_types, mlp_tp_comp)
    self.report('Testing distance computation for specific mlps with conn_coeffs=%s.'%(
                struct_coeffs))
    self._test_dist_comp_for_multiple_coeffs(mlps, dist_types, mislabel_coeffs,
                                             struct_coeffs, mlp_tp_comp)


class DistProdNNKernelTestCase(BaseTestClass):
  """ Unit tests for the Transport NNKernels. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(DistProdNNKernelTestCase, self).__init__(*args, **kwargs)
    self.non_assignment_penalty = 1
    cnn_layer_labels, label_mismatch_penalty = \
      otmann.get_cnn_layer_label_mismatch_penalties(self.non_assignment_penalty)
    self.all_layer_labels = cnn_layer_labels
    self.label_mismatch_penalty = label_mismatch_penalty
    self.tp_comp = otmann.OTMANNDistanceComputer(cnn_layer_labels,
                     label_mismatch_penalty,
                     self.non_assignment_penalty,
                     otmann.CNN_STRUCTURAL_PENALTY_GROUPS,
                     otmann.PATH_LENGTH_TYPES
                     )
    self.mislabel_coeffs = [2.0, 2.0, 1.0, 1.0, 1.0]
    self.struct_coeffs = [0.25, 0.5, 1.0, 2.0, 4.0]
    self.lp_betas = [1e-6] * len(self.struct_coeffs)
    self.emd_betas = [1] * len(self.struct_coeffs)
    self.scale = 1
    self.cnns = generate_cnn_architectures()

  def test_instantiation(self):
    """ Testing instantiation. """
    self.report('Testing instantiation of DistProdNNKernelTestCase and computation ' +
                'for specific networks.')
    dist_type_vals = ['lp', 'emd', 'lp-emd']
    all_kernels = []
    for dist_type in dist_type_vals:
      if dist_type == 'lp':
        betas = self.lp_betas
      elif dist_type == 'emd':
        betas = self.emd_betas
      else:
        betas = [j for i in zip(self.lp_betas, self.emd_betas) for j in i]
      tp_kernel = otmann.get_otmann_kernel_from_params('prod',
                    self.all_layer_labels, self.label_mismatch_penalty,
                    self.non_assignment_penalty,
                    otmann.CNN_STRUCTURAL_PENALTY_GROUPS,
                    otmann.PATH_LENGTH_TYPES,
                    self.mislabel_coeffs, self.struct_coeffs, dist_type,
                    betas, self.scale)
      cnn_K = tp_kernel(self.cnns)
      all_kernels.append(cnn_K)
      cnn_eig_vals, _ = np.linalg.eig(cnn_K)
      self.report('dist-type: %s, eigvals: %s.'%(dist_type,
                     get_list_of_floats_as_str(sorted(cnn_eig_vals))))
      self.report('%s transport kernel:\n%s'%(dist_type,
                  str(np.round(cnn_K, 3))), 'test_result')
      assert cnn_K.shape == (len(self.cnns), len(self.cnns))
      assert np.all(np.diag(cnn_K) == 1)
    # Check if it is in fact the product
    if 'lp' in dist_type_vals and 'emd' in dist_type_vals and 'lp-emd' in dist_type_vals:
      lp_kernel = all_kernels[dist_type_vals.index('lp')]
      emd_kernel = all_kernels[dist_type_vals.index('emd')]
      lpemd_kernel = all_kernels[dist_type_vals.index('lp-emd')]
      assert np.linalg.norm(lpemd_kernel - lp_kernel * emd_kernel) < _TOL

  def test_kernel_computation(self):
    """ Testing computation of the lp distance. """
    self.report('Testing computed kernel values for specific cnns.')
    betas = [0.0001]
    scale = 2.1
    struct_coeffs = 1.0
    mislabel_coeffs = 1.0
    dist_type = 'lp'
    tp_comp = otmann.OTMANNDistanceComputer(self.all_layer_labels,
                self.label_mismatch_penalty, self.non_assignment_penalty,
                otmann.CNN_STRUCTURAL_PENALTY_GROUPS,
                otmann.PATH_LENGTH_TYPES,
                dflt_mislabel_coeffs=mislabel_coeffs,
                dflt_struct_coeffs=struct_coeffs,
                dflt_dist_type=dist_type)
    tp_kernel = otmann.get_otmann_kernel_from_params('prod',
                self.all_layer_labels, self.label_mismatch_penalty,
                self.non_assignment_penalty, otmann.CNN_STRUCTURAL_PENALTY_GROUPS,
                otmann.PATH_LENGTH_TYPES,
                mislabel_coeffs, struct_coeffs, dist_type, betas, scale)
    cnn_dists = tp_comp(self.cnns, self.cnns)
    cnn_kernel = tp_kernel(self.cnns)
    diff = np.linalg.norm(scale * np.exp(-cnn_dists[0] * betas[0]) - cnn_kernel)
    assert diff < np.linalg.norm(cnn_kernel) * 1e-6


class DistSumNNKernelTestCase(BaseTestClass):
  """ Unit tests for the Transport NNKernels. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(DistSumNNKernelTestCase, self).__init__(*args, **kwargs)
    self.non_assignment_penalty = 1
    mlp_layer_labels, label_mismatch_penalty = \
      otmann.get_mlp_layer_label_mismatch_penalties(self.non_assignment_penalty,
                                                            'reg')
    self.all_layer_labels = mlp_layer_labels
    self.label_mismatch_penalty = label_mismatch_penalty
    self.tp_comp = otmann.OTMANNDistanceComputer(mlp_layer_labels,
                     label_mismatch_penalty,
                     self.non_assignment_penalty,
                     otmann.MLP_STRUCTURAL_PENALTY_GROUPS,
                     otmann.PATH_LENGTH_TYPES
                     )
    self.mislabel_coeffs = [2.0, 2.0, 1.0, 1.0, 1.0]
    self.struct_coeffs = [0.25, 0.5, 1.0, 2.0, 4.0]
    self.lp_betas = [1e-6] * len(self.struct_coeffs)
    self.emd_betas = [1] * len(self.struct_coeffs)
    self.mlps = generate_mlp_architectures()

  def test_instantiation_and_computation(self):
    """ Testing instantiation. """
    self.report('Testing instantiation of DistSumNNKernelTestCase and computation ' +
                'for specific networks.')
    dist_type_vals = ['lp', 'emd', 'lp-emd']
    all_kernels = []
    for dist_type in dist_type_vals:
      if dist_type == 'lp':
        betas = self.lp_betas
        scales = [1]
      elif dist_type == 'emd':
        betas = self.emd_betas
        scales = [1]
      else:
        betas = [j for i in zip(self.lp_betas, self.emd_betas) for j in i]
        scales = [1, 1]
      tp_kernel = otmann.get_otmann_kernel_from_params('sum',
                    self.all_layer_labels, self.label_mismatch_penalty,
                    self.non_assignment_penalty,
                    otmann.MLP_STRUCTURAL_PENALTY_GROUPS,
                    otmann.PATH_LENGTH_TYPES,
                    self.mislabel_coeffs, self.struct_coeffs, dist_type, betas, scales)
      nn_K = tp_kernel(self.mlps)
      nn_eig_vals, _ = np.linalg.eig(nn_K)
      self.report('dist-type: %s, eigvals: %s.'%(dist_type,
                     get_list_of_floats_as_str(sorted(nn_eig_vals))))
      assert nn_K.shape == (len(self.mlps), len(self.mlps))
      self.report('%s transport kernel:\n%s'%(dist_type,
                  str(np.round(nn_K, 3))), 'test_result')
      assert np.all(np.diag(nn_K) == sum(scales))
      all_kernels.append(nn_K)
    # Check if it is in fact the sum
    if 'lp' in dist_type_vals and 'emd' in dist_type_vals and 'lp-emd' in dist_type_vals:
      lp_kernel = all_kernels[dist_type_vals.index('lp')]
      emd_kernel = all_kernels[dist_type_vals.index('emd')]
      lpemd_kernel = all_kernels[dist_type_vals.index('lp-emd')]
      assert np.linalg.norm(lpemd_kernel - lp_kernel - emd_kernel) < _TOL

  def test_sum_product_equivalence(self):
    """ Unit-test for testing that both kernels compute the same thing in certain cases.
    """
    dist_type_vals = ['lp', 'emd']
    for dist_type in dist_type_vals:
      if dist_type == 'lp':
        betas = self.lp_betas
        scales = [1]
      elif dist_type == 'emd':
        betas = self.emd_betas
        scales = [1]
      sum_kernel = otmann.get_otmann_kernel_from_params('sum',
                     self.all_layer_labels, self.label_mismatch_penalty,
                     self.non_assignment_penalty,
                     otmann.MLP_STRUCTURAL_PENALTY_GROUPS,
                     otmann.PATH_LENGTH_TYPES,
                     self.mislabel_coeffs, self.struct_coeffs, dist_type, betas, scales)
      prod_kernel = otmann.get_otmann_kernel_from_params('prod',
                      self.all_layer_labels, self.label_mismatch_penalty,
                      self.non_assignment_penalty,
                      otmann.MLP_STRUCTURAL_PENALTY_GROUPS,
                      otmann.PATH_LENGTH_TYPES,
                      self.mislabel_coeffs, self.struct_coeffs, dist_type, betas, scales)
      sum_nn_K = sum_kernel(self.mlps)
      prod_nn_K = prod_kernel(self.mlps)
      assert np.linalg.norm(sum_nn_K - prod_nn_K) < _TOL


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for multi-objective optimisation using GP Bandits.
  -- kandasamy@cs.cmu.edu
"""
import unittest

from . import multiobjective_gp_bandit
from .unittest_euclidean_random_multiobjective_optimiser import \
  EuclideanMultiObjectiveOptimiserBaseTestCase
from ..utils.base_test_class import BaseTestClass, execute_tests


@unittest.skip
class EuclideanRandomMultiObjectiveOptimiserTestCase(
  EuclideanMultiObjectiveOptimiserBaseTestCase, BaseTestClass):
  """ Unit-tests for random multi-objective optimisation. """

  @classmethod
  def _child_instantiate_optimiser(cls, multi_func_caller, worker_manager, options,
                                   reporter):
    """ Instantiate optimiser. """
    return multiobjective_gp_bandit.EuclideanMultiObjectiveGPBandit(multi_func_caller,
             worker_manager, is_mf=False, options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, multi_func_caller, worker_manager, max_capital, mode,
                    *args, **kwargs):
    """ Runs multi-objective optimiser. """
    return multiobjective_gp_bandit.multiobjective_gpb_from_multi_func_caller(
             multi_func_caller, worker_manager, max_capital, is_mf=False,
             mode=mode, *args, **kwargs)


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for discrete distributions.
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import

# pylint: disable=invalid-name
# pylint: disable=no-self-use

import numpy as np

# Local imports
from .discrete import Bernoulli, Binomial, DiscreteUniform
from ..utils.base_test_class import BaseTestClass, execute_tests

class DiscreteDistributionsTestCase(BaseTestClass):
  """ Unit tests for distributions in discrete.py """

  def setUp(self):
    """ Sets up unit tests. """
    self.size = (10000, 10000)
    self.threshold = 0.01

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == samples.shape

  def _compute_mean(self, samples):
    """ Computes Mean """
    return np.mean(samples)

  def _compute_variance(self, samples):
    """ Computes Variance """
    return np.var(samples)

  def test_random_sampling_bernoulli(self):
    """ Tests random sampling from Bernoulli distribution """
    self.report('Test random sampling of Bernoulli Distribution.')
    p = 0.6
    dist = Bernoulli(p)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  def test_random_sampling_binomial(self):
    """ Tests random sampling from Binomial distribution """
    self.report('Test random sampling of Binomial Distribution.')
    n = 10
    p = 0.6
    dist = Binomial(n, p)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  def test_random_sampling_discreteuniform(self):
    """ Tests random sampling from Discrete Uniform distribution """
    self.report('Test random sampling of Discrete Uniform Distribution.')
    lower = -5
    upper = 5
    dist = DiscreteUniform(lower, upper)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()




"""
  Class for abstract probability distribution
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
# pylint: disable=invalid-name
# pylint: disable=abstract-method
# pylint: disable=relative-import

import numpy as np

# Local imports
from .distribution import Distribution
from ..sampling.slice import Slice
from ..sampling.nuts import NoUTurnSamplerDA as NUTS, LeapFrog
from ..sampling.metropolis import Metropolis, BinaryMetropolis

class Model(Distribution):
  """ Class for abstract distributions """

  def __init__(self, pdf, logp, grad_logp):
    super(Model, self).__init__()
    self._pdf = pdf
    self._logp = logp
    self._grad_logp = grad_logp
    self.dim = None

  def pdf(self, x):
    """ Returns pdf of distribution at x """
    return np.asscalar(self._pdf(x))

  def logp(self, x):
    """ Returns log of pdf at x """
    return self._logp(x)

  def grad_logp(self, x):
    """ Returns gradient of log pdf at x """
    return self._grad_logp(x)

  def draw_samples(self, method, size=None, init_sample=None, *args):
    """ Returns samples drawn from distribution. """
    if method == 'nuts':
      return self.draw_samples_nuts(size, init_sample, *args)
    elif method == 'slice':
      return self.draw_samples_slice(size, init_sample, *args)
    elif method == 'metropolis':
      return self.draw_samples_metropolis(size, init_sample, *args)
    elif method == 'binarymetropolis':
      return self.draw_samples_binary_metropolis(size, init_sample, *args)

  def draw_samples_slice(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Slice sampler"""
    sampler = Slice(self)
    return sampler.sample(init_sample, size, *args)

  def draw_samples_nuts(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using NUTS sampler"""
    sampler = NUTS(self, LeapFrog)
    return sampler.sample(init_sample, size, *args)

  def draw_samples_metropolis(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Metropolis sampler"""
    if hasattr(init_sample, '__len__'):
      self.dim = len(init_sample)
    else:
      self.dim = 1
    sampler = Metropolis(self, True, *args)
    return sampler.sample(init_sample, size)

  def draw_samples_binary_metropolis(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Binary Metropolis sampler"""
    sampler = BinaryMetropolis(self, True, *args)
    return sampler.sample(init_sample, size)



"""
  Unit tests for random multi-objective optimiser in Euclidean domains.
  -- bparia@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=abstract-class-little-used

import numpy as np
# Local imports
from ..exd.experiment_caller import EuclideanMultiFunctionCaller
from ..exd.worker_manager import SyntheticWorkerManager
from . import random_multiobjective_optimiser
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils import reporters


class EuclideanMultiObjectiveOptimiserBaseTestCase(object):
  """ A base test class for all optimisers on Euclidean domains. """
  # pylint: disable=no-member
  # pylint: disable=invalid-name

  def setUp(self):
    """ set up. """
    self.test_max_capital = 20
    self.multi_func_caller = EuclideanMultiFunctionCaller(
        funcs=[lambda x: (x[0] * x[1]), lambda x: (x[0] * np.sqrt(1-x[1]**2))],
        raw_domain=np.asarray([[0.1, 0.9]]*2),
        descr='simple_sphere',
        vectorised=False,
        to_normalise_domain=True,
    )
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')

  @classmethod
  def _child_instantiate_optimiser(cls, multi_func_caller, worker_manager, options,
                                   reporter):
    """ Instantiate a specific optimiser. """
    raise NotImplementedError('Implement in a child test class.')

  @classmethod
  def run_optimiser(cls, multi_func_caller, worker_manager, max_capital, mode,
                    *args, **kwargs):
    """ Runs optimiser. """
    raise NotImplementedError('Implement in a child test class.')

  def test_instantiation(self):
    """ Tests instantiation of the optimiser. """
    optimiser = self._child_instantiate_optimiser(
        self.multi_func_caller, self.worker_manager_3, options=None,
        reporter=reporters.get_reporter('silent'))
    self.report('Instantiated %s object.'%(type(optimiser)))
    for attr in dir(optimiser):
      if not attr.startswith('_'):
        self.report('optimiser.%s = %s'%(attr, str(getattr(optimiser, attr))),
                    'test_result')

  def _test_optimiser_results(self, pareto_vals, pareto_points, history):
    """ Tests optimiser results. """
    assert len(history.curr_pareto_vals) == len(history.curr_pareto_points)
    for val in pareto_vals:
      assert len(val) == self.multi_func_caller.num_funcs
    for val in pareto_points:
      assert len(val) == self.multi_func_caller.domain.get_dim()
    self.report('Pareto optimal points: %s.'%(pareto_points))
    self.report('Pareto optimal values: %s.'%(pareto_vals))

  def test_optimisation_single(self):
    """ Test optimisation with a single worker. """
    self.report('Testing %s with one worker.'%(type(self)))
    pareto_vals, pareto_points, history = self.run_optimiser(
        self.multi_func_caller, self.worker_manager_1, self.test_max_capital, 'asy')
    self._test_optimiser_results(pareto_vals, pareto_points, history)
    self.report('')
    return pareto_vals, pareto_points, history

  def test_optimisation_asynchronous(self):
    """ Testing random optimiser with three asynchronous workers. """
    self.report('Testing %s with three asynchronous workers.'%(type(self)))
    pareto_vals, pareto_points, history = self.run_optimiser(
        self.multi_func_caller, self.worker_manager_3, self.test_max_capital, 'asy')
    self._test_optimiser_results(pareto_vals, pareto_points, history)
    self.report('')
    return pareto_vals, pareto_points, history


class EuclideanRandomMultiObjectiveOptimiserTestCase(
  EuclideanMultiObjectiveOptimiserBaseTestCase, BaseTestClass):
  """ Unit-tests for random multi-objective optimisation. """

  @classmethod
  def _child_instantiate_optimiser(cls, multi_func_caller, worker_manager, options,
                                   reporter):
    """ Instantiate optimiser. """
    return random_multiobjective_optimiser.EuclideanRandomMultiObjectiveOptimiser(
             multi_func_caller, worker_manager, options, reporter)

  @classmethod
  def run_optimiser(cls, multi_func_caller, worker_manager, max_capital, mode,
                    *args, **kwargs):
    """ Runs multi-objective optimiser. """
    rmoo = random_multiobjective_optimiser
    return rmoo.random_multiobjective_optimisation_from_multi_func_caller(
             multi_func_caller, worker_manager, max_capital, mode, *args, **kwargs)


if __name__ == '__main__':
  execute_tests()




"""
  Implements some example neural networks. We will mostly use them to instantiate the
  chains in GA and/or Simulated annealing.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=abstract-class-not-used
# pylint: disable=no-name-in-module

import numpy as np
from scipy.sparse import dok_matrix
# local imports
from .neural_network import ConvNeuralNetwork, MultiLayerPerceptron, \
                            get_cnn_layer_labels,\
                            get_mlp_layer_labels, is_a_conv_layer_label
from .nn_modifiers import NNModifier
from ..utils.general_utils import get_dok_mat_with_set_coords


# CNNs ===================================================================================
def get_feedforward_adj_mat(num_layers):
  """ Returns an adjacency matrix for a feed forward network. """
  ret = dok_matrix((num_layers, num_layers))
  for i in range(num_layers-1):
    ret[i, i+1] = 1
  return ret

def get_vgg_net(num_conv_layers_per_block=4, cnn_layer_labels=None):
  """ Returns a VGG net. """
  cnn_layer_labels = cnn_layer_labels if cnn_layer_labels is not None else \
                     get_cnn_layer_labels()
  layer_labels = ['ip', 'conv3', 'conv3', 'max-pool', 'conv3', 'conv3', 'max-pool']
  num_filters_each_layer = [None, 64, 64, None, 128, 128, None]
  # Now create the blocks
  block_filter_sizes = [128, 256, 512]
  for bfs in block_filter_sizes:
    layer_labels.extend(['conv3' for _ in range(num_conv_layers_per_block)] +
                        ['max-pool'])
    num_filters_each_layer.extend([bfs] * num_conv_layers_per_block + [None])
  layer_labels.extend(['fc', 'fc', 'fc', 'softmax', 'op'])
  num_filters_each_layer.extend([128, 256, 512, None, None])
  num_layers = len(layer_labels)
  # Construct the connectivity matrix
  conn_mat = get_feedforward_adj_mat(num_layers)
  strides = [(1 if is_a_conv_layer_label(ll) else None) for ll in layer_labels]
  vgg = ConvNeuralNetwork(layer_labels, conn_mat, num_filters_each_layer, strides,
                          cnn_layer_labels)
  return vgg


def _get_multidepth_cnn_eg12_common():
  """ A network with 2 softmax layers mostly for debugging common operations. """
  cnn_layer_labels = get_cnn_layer_labels()
  layer_labels = ['ip', 'op', 'softmax', 'fc', 'softmax', 'fc', 'conv5', 'avg-pool',
                  'max-pool', 'conv3', 'conv3', 'max-pool', 'max-pool', 'conv3', 'conv7']
  num_filters_each_layer = [None, None, None, 64, None, 64, 128, None, None, 64,
                            64, None, None, 128, 64]
  edges = [(0, 14), (14, 6), (14, 9), (14, 10), (6, 7), (7, 3), (3, 2), (2, 1), (9, 8),
           (8, 5), (5, 4), (4, 1), (10, 11), (11, 13), (13, 12), (12, 5)]
  strides = [(1 if is_a_conv_layer_label(ll) else None) for ll in layer_labels]
  return layer_labels, edges, num_filters_each_layer, cnn_layer_labels, strides


def get_multidepth_cnn_eg1():
  """ A network with 2 softmax layers mostly for debugging. """
  layer_labels, edges, num_filters_each_layer, cnn_layer_labels, strides = \
    _get_multidepth_cnn_eg12_common()
  edges.append((3, 4))
  conn_mat = get_dok_mat_with_set_coords(len(layer_labels), edges)
  strides[9] = 2
  return ConvNeuralNetwork(layer_labels, conn_mat, num_filters_each_layer, strides,
                           cnn_layer_labels)


def get_multidepth_cnn_eg2():
  """ A network with 2 softmax layers mostly for debugging. """
  layer_labels, edges, num_filters_each_layer, cnn_layer_labels, strides = \
    _get_multidepth_cnn_eg12_common()
  edges.append((6, 8))
  conn_mat = get_dok_mat_with_set_coords(len(layer_labels), edges)
  strides[9] = 2
  strides[6] = 2
  return ConvNeuralNetwork(layer_labels, conn_mat, num_filters_each_layer, strides,
                           cnn_layer_labels)


def _dflt_unit_sizes(num_blocks):
  """ Returns a default set of values for the unit sizes. """
  ret = []
  unit_size = 64
  for _ in range(num_blocks):
    ret.append(unit_size)
    unit_size = min(unit_size*2, 512)
  return ret


def _get_blocked_cnn_params(num_blocks, num_layers_per_block, block_layer_type,
  num_fc_layers, num_conv_filters_in_layers=None, num_fc_nodes_in_layers=None,
  cnn_layer_labels=None):
  """ Returns parameters for a blocked CNN.
    num_blocks: # blocks, i.e. the number of repeated convolutional layers.
    num_layers_per_block: # layers per block.
    num_fc_layers: # fully connected layers.
    num_conv_filters_in_layers: # filters in the layers for each block.
    num_fc_nodes_in_layers: sizes of the fc layers.
    cnn_layer_labels: Labels for all layer types in a CNN.
  """
  layer_labels = ['ip', 'conv7', 'max-pool']
  num_filters_each_layer = [None, 64, None]
  strides = [None, 1, None]
  num_conv_filters_in_layers = num_conv_filters_in_layers if num_conv_filters_in_layers \
    is not None else _dflt_unit_sizes(num_blocks)
  num_fc_nodes_in_layers = num_fc_nodes_in_layers if num_fc_nodes_in_layers is not None \
    else [2 * num_conv_filters_in_layers[-1]] * num_fc_layers
  cnn_layer_labels = cnn_layer_labels if cnn_layer_labels is not None else \
                     get_cnn_layer_labels()
  # Construct blocks
  for block_idx in range(num_blocks):
    layer_labels.extend([block_layer_type for _ in range(num_layers_per_block)])
    num_filters_each_layer.extend([num_conv_filters_in_layers[block_idx]] *
                         num_layers_per_block)
    strides.extend([2] + [1] * (num_layers_per_block - 1))
  # Pooling layer after the blocks
  layer_labels.append('avg-pool')
  num_filters_each_layer.append(None)
  strides.append(None)
  # Add FC layers
  layer_labels.extend(['fc' for _ in range(num_fc_layers)] + ['softmax', 'op'])
  num_filters_each_layer.extend(num_fc_nodes_in_layers + [None, None])
  strides.extend([None] * (num_fc_layers + 2))
  # Construct the connectivity matrix
  num_layers = len(layer_labels)
  conn_mat = get_feedforward_adj_mat(num_layers)
  return layer_labels, conn_mat, num_filters_each_layer, cnn_layer_labels, strides


def get_blocked_cnn(num_blocks, num_conv_layers_per_block, num_fc_layers,
                    num_conv_filters_in_layers=None,
                    num_fc_nodes_in_layers=None, cnn_layer_labels=None):
  """ Returns a blocked CNN.
    num_blocks: # blocks, i.e. the number of repeated convolutional layers.
    num_conv_layers_per_block: # convolutaional layers per block.
    num_fc_layers: # fully connected layers.
    num_conv_filters_in_layers: # filters in the layers for each block.
    num_fc_nodes_in_layers: sizes of the fc layers.
    cnn_layer_labels: Labels for all layer types in a CNN.
  """
  layer_labels, conn_mat, num_filters_each_layer, cnn_layer_labels, strides = \
    _get_blocked_cnn_params(num_blocks, num_conv_layers_per_block, 'conv3', num_fc_layers,
      num_conv_filters_in_layers, num_fc_nodes_in_layers, cnn_layer_labels)
  return ConvNeuralNetwork(layer_labels, conn_mat, num_filters_each_layer, strides,
                           cnn_layer_labels)


def get_resnet_cnn(num_res_blocks, num_conv_layers_per_block, num_fc_layers,
                   num_conv_filters_in_layers=None,
                   num_fc_nodes_in_layers=None, cnn_layer_labels=None):
  """ Returns a Resnet CNN.
      num_layers_to_skip: The number of layers to skip when adding skip connections.
      see get_blocked_cnn for other arguments.
  """
  layer_labels, conn_mat, num_filters_each_layer, cnn_layer_labels, strides = \
    _get_blocked_cnn_params(num_res_blocks, num_conv_layers_per_block, 'res3',
      num_fc_layers, num_conv_filters_in_layers, num_fc_nodes_in_layers, cnn_layer_labels)
  return ConvNeuralNetwork(layer_labels, conn_mat, num_filters_each_layer, strides,
                           cnn_layer_labels)


#MLPs ===================================================================================
def get_blocked_mlp(class_or_reg, num_blocks, num_layers_per_block=None,
                    num_units_in_each_layer=None):
  """ Creates a blocked MLP. """
  # Create rectifiers and sigmoids
  rectifiers = ['relu', 'elu', 'crelu', 'leaky-relu', 'softplus']
  sigmoids = ['logistic', 'tanh']
  obtl_label = 'linear' if class_or_reg == 'reg' else 'softmax'
  np.random.shuffle(rectifiers)
  np.random.shuffle(sigmoids)
  rect_count = 0
  sig_count = 0
  # Preprocess args
  # Create the network
  layer_labels = ['ip']
  num_units_in_each_layer = num_units_in_each_layer if num_units_in_each_layer \
                            is not None else _dflt_unit_sizes(num_blocks)
  # Construct blocks
  for block_idx in range(num_blocks):
    if block_idx % 2 == 0:
      layer_type = rectifiers[rect_count%len(rectifiers)]
      rect_count += 1
    else:
      layer_type = sigmoids[sig_count%len(sigmoids)]
      sig_count += 1
    layer_labels.extend([layer_type] * num_layers_per_block)
    num_units_in_each_layer.extend([num_units_in_each_layer[block_idx]] *
                                   num_layers_per_block)
  # A linear layer at the end of the block
  layer_labels.extend([obtl_label, 'op'])
  num_units_in_each_layer.extend([None, None])
  conn_mat = get_feedforward_adj_mat(len(layer_labels))
  all_layer_labels = get_mlp_layer_labels(class_or_reg)
  return MultiLayerPerceptron(class_or_reg, layer_labels, conn_mat,
                              num_units_in_each_layer, all_layer_labels)


def _get_multidepth_mlp_eg12_common(class_or_reg):
  """ A network with 2 linear layers mostly for debugging common operations. """
  mlp_layer_labels = get_mlp_layer_labels(class_or_reg)
  obtl_label = 'linear' if class_or_reg == 'reg' else 'softmax'
  layer_labels = ['ip', obtl_label, 'op', 'tanh', 'relu', 'leaky-relu', 'logistic',
                  'logistic', 'elu', obtl_label]
  num_units_in_each_layer = [None, None, None, 64, 64, 128, 256, 64, 512, None]
  edges = [(0, 1), (0, 3), (0, 4), (3, 5), (4, 7), (5, 6), (7, 8), (3, 8), (7, 6), \
           (6, 9), (8, 9), (1, 2), (9, 2)]
  return layer_labels, num_units_in_each_layer, edges, mlp_layer_labels

def get_multidepth_mlp_eg1(class_or_reg):
  """ Multi depth MLP eg 1. """
  layer_labels, num_units_in_each_layer, edges, mlp_layer_labels = \
    _get_multidepth_mlp_eg12_common(class_or_reg)
  conn_mat = get_dok_mat_with_set_coords(len(layer_labels), edges)
  return MultiLayerPerceptron(class_or_reg, layer_labels, conn_mat,
                              num_units_in_each_layer, mlp_layer_labels)

def get_multidepth_mlp_eg2(class_or_reg):
  """ Multi depth MLP eg 2. """
  layer_labels, num_units_in_each_layer, edges, mlp_layer_labels = \
    _get_multidepth_mlp_eg12_common(class_or_reg)
  edges.append((4, 1))
  conn_mat = get_dok_mat_with_set_coords(len(layer_labels), edges)
  return MultiLayerPerceptron(class_or_reg, layer_labels, conn_mat,
                              num_units_in_each_layer, mlp_layer_labels)


# Some functions to generate neural network architectures ---------------------------
def generate_cnn_architectures():
  # pylint: disable=bad-whitespace
  """ Generates 4 neural networks. """
  all_layer_label_classes = get_cnn_layer_labels()
  # Network 1
  layer_labels = ['ip', 'op', 'conv3', 'fc', 'conv3', 'conv3', 'conv3', 'softmax',
                  'max-pool']
  num_filters_each_layer = [None, None, 32, 16, 16, 8, 8, None, None]
  A = get_dok_mat_with_set_coords(9,
    [(0,4), (3,7), (4,5), (4,6), (5,2), (6,2), (7,1), (2, 8), (8, 3)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'max-pool']
             else 1 for ll in layer_labels]
  cnn_1 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)
  # Network 2
  layer_labels = ['ip', 'conv3', 'conv3', 'conv3', 'fc', 'softmax', 'op', 'max-pool']
  num_filters_each_layer = [None, 16, 16, 32, 16, None, None, None]
  A = get_dok_mat_with_set_coords(8, [(0,1), (1,2), (2,3), (4,5), (5,6), (3, 7),
                                      (7, 4)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'max-pool']
             else 1 for ll in layer_labels]
  cnn_2 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)
  # Network 3
  layer_labels = ['ip', 'conv3', 'conv3', 'conv5', 'conv3', 'max-pool',
                  'fc', 'softmax', 'op']
  num_filters_each_layer = [None, 16, 16, 16, 32, None, 32, None, None]
  A = get_dok_mat_with_set_coords(9, [(0,1), (1,2), (1,3), (1,4), (2,5), (3,5), (4,6),
                                       (5,6), (6,7), (7,8)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'max-pool']
             else 1 for ll in layer_labels]
  strides[4] = 2
  cnn_3 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)
  # Network 4
  layer_labels = ['ip', 'conv3', 'conv3', 'conv5', 'conv3', 'avg-pool', 'conv5', 'fc',
                  'softmax', 'op']
  num_filters_each_layer = [None, 16, 16, 16, 32, None, 32, 32, None, None]
  A = get_dok_mat_with_set_coords(10, [(0,1), (1,2), (1,3), (1,4), (2,5), (3,5), (4,6),
                                        (5,7), (6,7), (7,8), (8,9)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'max-pool', 'avg-pool']
             else 1 for ll in layer_labels]
  strides[4] = 2
  cnn_4 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)
  # Network 5
  layer_labels = ['ip', 'conv3', 'conv3', 'conv5', 'conv5', 'avg-pool',
                  'fc', 'softmax', 'op', 'conv3']
  num_filters_each_layer = [None, 16, 16, 16, 32, None, 32, None, None, 16]
  A = get_dok_mat_with_set_coords(10, [(0,1), (1,2), (1,3), (2,5), (3,5), (4,6),
                                       (5,6), (6,7), (7,8), (0, 9), (9, 4)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'avg-pool']
             else 1 for ll in layer_labels]
  strides[4] = 2
  cnn_5 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)
  # Network 6
  layer_labels = ['ip', 'conv3', 'conv3', 'conv3', 'fc', 'fc', 'op', 'max-pool',
                  'fc', 'softmax']
  num_filters_each_layer = [None, 16, 16, 32, 32, 32, None, None, 32, None]
  A = get_dok_mat_with_set_coords(10, [(0,1), (1,2), (2,3), (4,5), (5,9), (3,7),
                                      (7,4), (4,8), (8,9), (9,6)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'max-pool']
             else 1 for ll in layer_labels]
  cnn_6 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)
  # Network 7 - Two softmax layers both pointing to output
  layer_labels = ['ip', 'conv3', 'conv3', 'conv3', 'fc', 'fc', 'op', 'max-pool',
                  'fc', 'softmax', 'softmax']
  num_filters_each_layer = [None, 16, 16, 32, 32, 32, None, None, 32, None, None]
  A = get_dok_mat_with_set_coords(11, [(0,1), (1,2), (2,3), (4,5), (5,9), (3,7),
                                      (7,4), (4,8), (8,9), (9,6), (8,10), (10,6)])
  strides = [None if ll in ['ip', 'op', 'fc', 'softmax', 'max-pool']
             else 1 for ll in layer_labels]
  cnn_7 = ConvNeuralNetwork(layer_labels, A,
            num_filters_each_layer, strides, all_layer_label_classes)

  return [cnn_1, cnn_2, cnn_3, cnn_4, cnn_5, cnn_6, cnn_7,
          get_vgg_net(2),
          get_blocked_cnn(3, 4, 1),
          get_resnet_cnn(3, 2, 1),
          get_multidepth_cnn_eg1(),
          get_multidepth_cnn_eg2(),
         ]


def generate_mlp_architectures(class_or_reg='reg'):
  """ pylint: disable=bad-whitespace. """
  # pylint: disable=bad-whitespace
  all_layer_label_classes = get_mlp_layer_labels(class_or_reg)
  last_layer_label = 'linear' if class_or_reg == 'reg' else 'softmax'
  # Network 1
  layer_labels = ['ip', 'op', 'tanh', 'logistic', 'softplus', 'relu', 'elu',
                  last_layer_label]
  num_units_each_layer = [None, None, 32, 64, 16, 8, 8, None]
  A = get_dok_mat_with_set_coords(8,
    [(0,4), (2,3), (3,7), (4,5), (4,6), (5,2), (6,2), (7,1)])
  mlp_1 = MultiLayerPerceptron(class_or_reg, layer_labels, A,
            num_units_each_layer, all_layer_label_classes)
  # Network 2
  layer_labels = ['ip', 'softplus', 'elu', 'tanh', 'logistic', last_layer_label, 'op']
  num_units_each_layer = [None, 16, 16, 32, 64, None, None]
  A = get_dok_mat_with_set_coords(7, [(0,1), (1,2), (2,3), (3,4), (4,5), (5,6)])
  mlp_2 = MultiLayerPerceptron(class_or_reg, layer_labels, A,
            num_units_each_layer, all_layer_label_classes)
  # Network 3
  layer_labels = ['ip', 'tanh', 'logistic', 'logistic', 'tanh', 'elu', 'relu',
                  last_layer_label, 'op']
  num_units_each_layer = [None, 8, 8, 8, 8, 16, 16, None, None]
  A = get_dok_mat_with_set_coords(9, [(0, 1), (0, 2), (1, 3), (2, 4), (3, 5), (3, 6),
                                      (4, 5), (4, 6), (5, 7), (6, 7), (7, 8)])
  mlp_3 = MultiLayerPerceptron(class_or_reg, layer_labels, A,
            num_units_each_layer, all_layer_label_classes)
  # Network 4
  layer_labels = ['ip', 'logistic', 'relu', 'softplus', last_layer_label, 'op', 'relu',
                  'tanh']
  num_units_each_layer = [None, 8, 8, 16, None, None, 16, 8]
  A = get_dok_mat_with_set_coords(8, [(0, 1), (0, 7), (1, 2), (2, 3), (2, 6), (7, 6),
                                      (7, 3), (3, 4), (6, 4), (4, 5)])
  mlp_4 = MultiLayerPerceptron(class_or_reg, layer_labels, A,
            num_units_each_layer, all_layer_label_classes)
  return [mlp_1, mlp_2, mlp_3, mlp_4,
          get_blocked_mlp(class_or_reg, 4, 3),
          get_blocked_mlp(class_or_reg, 8, 2),
          get_multidepth_mlp_eg1(class_or_reg),
          get_multidepth_mlp_eg2(class_or_reg),
         ]


def generate_many_neural_networks(nn_type, num_nns=200, initial_nns=None,
                                  nn_constraint_checker=None):
  """ Generates a large number of MLPs. """
  if initial_nns is None:
    if nn_type == 'cnn':
      initial_nns = generate_cnn_architectures()
    elif nn_type == 'mlp-reg':
      initial_nns = generate_mlp_architectures('reg')
    elif nn_type == 'mlp-class':
      initial_nns = generate_mlp_architectures('class')
    else:
      raise ValueError('Unknown nn_type: %s.'%(nn_type))
  num_steps_probs = np.array([0.5, 0.25, 0.125, 0.075, 0.05])
  num_steps_probs = num_steps_probs / num_steps_probs.sum()
  num_iters = 5
  num_nns_per_iter = int(num_nns / float(num_iters))
  # Get modifier
  modifier = NNModifier(nn_constraint_checker)
  curr_nns = initial_nns
  ret = [elem for elem in initial_nns]
  while len(ret) <= num_nns:
    curr_nns = modifier(curr_nns, num_nns_per_iter, num_steps_probs)
    ret.extend(curr_nns)
  np.random.shuffle(ret)
  return ret[:num_nns]

